import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
from myload import load

dummy = go.Figure()
dummy.write_image("dummy.pdf", format="pdf")

la_list = [0.0, 0.02, 0.05, 0.1, 0.15]
lu_list = [0.0, 0.005, 0.01]


def compress_legend(fig):
    group1_base, group2_base = fig.data[0].name.split(",")
    lines_marker_name = []
    for i, trace in enumerate(fig.data):
        part1, part2 = trace.name.split(",")
        if part1 == group1_base:
            lines_marker_name.append(
                dict(
                    line=trace.line.to_plotly_json(),
                    marker=trace.marker.to_plotly_json(),
                    mode=trace.mode,
                    name=part2.lstrip(" "),
                )
            )
        if part2 == group2_base:
            trace["legend"] = "legend2"
            trace["legendgroup"] = "part2"
            trace["name"] = part1
        else:
            trace["name"] = ''
            trace["showlegend"] = False
    for lmn in lines_marker_name:
        lmn["line"]["color"] = "black"
        lmn["marker"]["color"] = "black"
        lmn["legend"] = "legend1"
        lmn["legendgroup"] = "part1"
        fig.add_trace(go.Scatter(y=[None], **lmn))
    fig.update_layout(
        legend=dict(
            title_text="$\\lambda_U$",
            x=0.7,
            y= 0.9,
        ),
        legend2=dict(
            title_text="$\\lambda_A$",
            x=0.4,
            y= 0.9,
        ),
    )


for data, init, clip in [
    ("Kd32", "min31", "clip2x2"),
    ("CA3", "min27", "clip3-100"),
]:
    out = []
    for la in la_list:
        #    for la in [0, 20, 40, 60, 80, 100]:
        for lu in lu_list:
            path = f"outputs/{data}/miniature/{init}/{clip}/su{lu}/sa{la}/pos0"
            for stage in range(30):
                try:
                    df = load(path, stage, df_only=True)
                    out.append((lu, la, stage, np.count_nonzero(df.kind == "cell")))
                except FileNotFoundError:
                    pass
    lu, la, stage, num = zip(*out)
    df = pd.DataFrame(dict(lu=lu, la=la, stage=stage, num=num))
    fig = px.line(
        df,
        x="stage",
        y="num",
        color="la",
        symbol="lu",
        template="none",
    )
    compress_legend(fig)
    fig.update_xaxes(
        autorange=False,
        range=(-0.5, 30.5),
        title_text="epoch",
    )
    fig.update_yaxes(
        autorange=False,
        range=(0, 1.05 * df.num.max()),
        title_text="# candidates",
    )
    fig.update_layout(
        font_size=11,
        width=300,
        height=300,
        margin=dict(l=35, r=5, t=5, b=25),
    )
    pio.full_figure_for_development(fig, warn=False)
    print(fig.data)
    print(fig.layout)
    fig.write_image(f"figs/{data}-num.pdf", engine="kaleido")



from logging import getLogger
from pathlib import Path

import numpy as np

from ..io import (
    apply_mask,
    load_imgs,
    try_load,
)

# from hotaru.io.movie import gen_normalize_movie
from ..io.plot import (
    plot_peak_stats,
    plot_seg,
    plot_seg_max,
    plot_simgs,
    plot_spike,
)

logger = getLogger(__name__)


def plotter(cfg):
    def load(name):
        odir = Path(cfg.outputs.dir)
        path = cfg.outputs[name]
        fdir = odir / path.dir
        files = [fdir / file.format(stage=stage) for file in path.files]
        out = try_load(files)
        if out is None or np.any([o is None for o in out]):
            raise RuntimeError(f"load failed: {files}")
        else:
            logger.info("loaded:" + " %s" * len(files), *files)
        return out

    fig_dir = Path(cfg.outputs.dir) / cfg.outputs.figs.dir
    fig_dir.mkdir(parents=True, exist_ok=True)
    time = cfg.plot.time

    imgs, hz = load_imgs(**cfg.data.imgs)
    imgs, mask = apply_mask(imgs, **cfg.data.mask)

    stage = cfg.plot.stage

    if stage == 0:
        stats, *simgs = load("stats")
        findval = load("find")
        footprints, peaks = load("make")
    else:
        footprints, peaks = load("spatial")

    # gen_normalize_movie("test.mp4", data)
    # plot_gl(data, radius, [100, 200, 300], scale=0.3).write_image(fig_dir / "gl.pdf")
    if stage == 0:
        plot_simgs(simgs).write_image(fig_dir / "stats.pdf")
        plot_peak_stats(peaks, findval).write_image(fig_dir / f"{stage:03d}peaks.pdf")
    else:
        plot_peak_stats(peaks).write_image(fig_dir / f"{stage:03d}peaks.pdf")
    plot_seg_max(footprints, peaks, **cfg.plot.seg_max).write_image(fig_dir / f"{stage:03d}max.pdf")
    plot_seg(footprints, peaks, 10).write_image(fig_dir / f"{stage:03d}seg.pdf")

    spikes, background = load("temporal")
    #stats = load("spike")
    diff = spikes.shape[1] - imgs.shape[0]

    plot_spike(spikes, hz, diff, time).write_image(fig_dir / f"{stage:03d}spike.pdf")



from logging import getLogger

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf

from ..utils import (
    from_tf,
    get_gpu_env,
)

logger = getLogger(__name__)


def prepare_matrix(data, y, trans, env, factor, prefetch):
    nt = data.nt
    ns = data.ns
    nk, ny = y.shape

    env = get_gpu_env(env)
    nd = env.num_devices
    sharding = env.sharding((nd, 1))
    batch = env.batch(float(factor) * nk * ns, nt)
    logger.info("prepare: nt=%d nk=%d ns=%d batch=%d", nt, nk, ns, batch)

    nkmod = nd * ((nk + nd - 1) // nd)
    y = np.pad(y, ((0, nkmod - nk), (0, 0)))
    yval = jax.device_put(y, sharding)

    yavg = yval.mean(axis=1)
    ydif = yval - yavg[:, jnp.newaxis]

    ycov = ydif @ ydif.T
    yout = jnp.outer(yavg, yavg)

    logger.info("%s: %s %s %d", "pbar", "start", "prepare", nt)
    ydot = matmul_batch(ydif, data, trans, sharding, batch, prefetch)
    logger.info("%s: %s", "pbar", "close")

    return ycov[:nk, :nk], yout[:nk, :nk], ydot[:nk]


def matmul_batch(x, y, trans, sharding, batch, prefetch):
    if trans:

        @jax.jit
        def calc(out, x, t, yt):
            return out.at[:, t].set(x @ yt.T)

        nk, ns = x.shape
        nt = y.nt
        ntmod = batch * ((nt + batch - 1) // batch)
        out = jax.device_put(jnp.empty((nk, ntmod)), sharding)
    else:

        @jax.jit
        def calc(out, x, t, yt):
            return out + x[:, t] @ yt

        nk, nt = x.shape
        ns = y.ns
        out = jax.device_put(jnp.zeros((nk, ns)), sharding)

    dataset = tf.data.Dataset.from_generator(
        lambda: enumerate(y.data(mask_type=True)),
        output_signature=(
            tf.TensorSpec((), tf.int32),
            tf.TensorSpec((ns,), tf.float32),
        ),
    )
    dataset = dataset.batch(batch)
    dataset = dataset.prefetch(prefetch)

    end = 0
    for data in dataset:
        t, yt = tuple(jax.device_put(from_tf(d), sharding.replicate()) for d in data)
        out = calc(out, x, t, yt)
        start, end = end, end + batch
        n = batch if end < nt else nt - start
        logger.info("%s: %s %d", "pbar", "update", n)

    if trans:
        out = out[:, :nt]
    return out


def loss_fn(x, ycov, yout, ydot, nx, ny, bx, by):
    nn = nx * ny
    nm = nn + nx + ny
    xsum = x.sum(axis=1)
    xavg = xsum / nx
    xdif = x - xavg[:, jnp.newaxis]
    xcov = xdif @ xdif.T
    xout = jnp.outer(xavg, xavg)
    diff = (ycov * xcov).sum() - 2 * (ydot * xdif).sum() + nn
    penalty = ny * bx * (yout * xcov).sum() + nx * by * (ycov * xout).sum()
    var = diff + penalty
    return (nm / 2) * (jnp.log(var) - jnp.log(nm)), jnp.sqrt(var / nm)



from .dynamics import (
    SpikeToCalcium,
    CalciumToSpike,
    get_dynamics,
    get_rdynamics,
)
from .evaluate import (
    fix_kind,
    evaluate,
)

__all__ = [
    "CalciumToSpike",
    "SpikeToCalcium",
    "get_dynamics",
    "get_rdynamics",
    "fix_kind",
    "evaluate",
]



from collections import namedtuple
from logging import getLogger

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf

from ..utils import (
    from_tf,
    get_gpu_env,
)
from .neighbor import neighbor

logger = getLogger(__name__)

Stats = namedtuple("Stats", "avgx avgt std0 min0 max0 min1 max1")


def movie_stats(raw_imgs, mask=None, env=None, factor=1, prefetch=1):
    @jax.jit
    def update(avgt, sumi, sqi, sumn, sqn, cor, min0, max0, imin, imax, index, imgs):
        if mask is None:
            masked = imgs.reshape(batch, h * w)
        else:
            masked = imgs[:, mask]

        avgti = jnp.nanmean(masked, axis=1)
        avgt = avgt.at[index].set(avgti)

        diff = imgs - avgti[..., jnp.newaxis, jnp.newaxis]
        sumi += jnp.nansum(diff, axis=0)
        sqi += jnp.nansum(jnp.square(diff), axis=0)

        neig = neighbor(diff)
        sumn += jnp.nansum(neig, axis=0)
        sqn += jnp.nansum(jnp.square(neig), axis=0)

        cor += jnp.nansum(diff * neig, axis=0)

        min0 = jnp.minimum(min0, jnp.nanmin(imgs, axis=0))
        max0 = jnp.maximum(max0, jnp.nanmax(imgs, axis=0))
        imin = jnp.minimum(imin, jnp.nanmin(diff, axis=0))
        imax = jnp.maximum(imax, jnp.nanmax(diff, axis=0))

        return avgt, sumi, sqi, sumn, sqn, cor, min0, max0, imin, imax

    nt, h, w = raw_imgs.shape

    env = get_gpu_env(env)
    batch = env.batch(float(factor) * h * w, nt)
    nd = env.num_devices
    sharding = env.sharding((nd, 1))

    logger.info("stats batch: %d", batch)
    dataset = tf.data.Dataset.from_generator(
        lambda: zip(range(nt), raw_imgs),
        output_signature=(
            tf.TensorSpec((), tf.int32),
            tf.TensorSpec((h, w), tf.float32),
        ),
    )
    dataset = dataset.batch(batch)
    dataset = dataset.prefetch(prefetch)

    avgt = jnp.empty((nt + 1,))

    sumi = jnp.zeros((h, w))
    sqi = jnp.zeros((h, w))

    sumn = jnp.zeros((h, w))
    sqn = jnp.zeros((h, w))

    cor = jnp.zeros((h, w))

    min0 = jnp.full((h, w), jnp.inf)
    max0 = jnp.full((h, w), -jnp.inf)
    imin = jnp.full((h, w), jnp.inf)
    imax = jnp.full((h, w), -jnp.inf)

    avgt, sumi, sqi, sumn, sqn, cor, min0, max0, imin, imax = (
        jax.device_put(v, sharding) for v in [
            avgt, sumi, sqi, sumn, sqn, cor, min0, max0, imin, imax,
        ]
    )

    logger.info("%s: %s %s %d", "pbar", "start", "stats", nt)
    for data in dataset:
        data = (from_tf(v) for v in data)
        index, imgs = (jax.device_put(v) for v in data)

        count = index.size
        diff = batch - count
        if diff > 0:
            pad = (0, diff), (0, 0), (0, 0)
            index = jnp.pad(index, ((0, diff)), constant_values=-1)
            imgs = jnp.pad(imgs, pad, constant_values=jnp.nan)

        index = jax.device_put(index, sharding)
        imgs = jax.device_put(imgs, sharding)

        avgt, sumi, sqi, sumn, sqn, cor, min0, max0, imin, imax = update(
            avgt, sumi, sqi, sumn, sqn, cor, min0, max0, imin, imax, index, imgs,
        )
        logger.info("%s: %s %d", "pbar", "update", count)
    logger.info("%s: %s", "pbar", "close")

    avgt = avgt[:-1]
    avgx = sumi / nt
    varx = sqi / nt - jnp.square(avgx)
    if mask is None:
        varx_masked = varx.ravel()
    else:
        varx_masked = varx[mask]

    std0 = jnp.sqrt(varx_masked.mean())
    stdx = jnp.sqrt(varx)
    istd = stdx / std0

    avgn = sumn / nt
    stdn = jnp.sqrt(sqn / nt - jnp.square(avgn))
    icor = (cor / nt - avgx * avgn) / (stdx * stdn)

    imin = (imin - avgx) / std0
    imax = (imax - avgx) / std0

    if mask is not None:
        avgx.at[~mask].set(jnp.nan)
        imin.at[~mask].set(jnp.nan)
        imax.at[~mask].set(jnp.nan)
        istd.at[~mask].set(jnp.nan)
        icor.at[~mask].set(jnp.nan)

    stats = avgx, avgt, std0, min0.min(), max0.max(), imin.min(), imax.max()
    simgs = imax, istd, icor
    return Stats(*map(np.array, stats)), *map(np.array, simgs)



import dash.dcc as dcc
import dash.html as html
import dash_bootstrap_components as dbc
import plotly.graph_objects as go
from dash import (
    Input,
    Output,
    Patch,
    State,
    callback,
)
from omegaconf import OmegaConf

from ..main import (
    make,
    reduce,
)
from .graph import (
    img_fig,
    update_img_fig,
)
from .progress import Progress


def spike_fig():
    fig = go.Figure()
    return fig


def update_spike_fig(fig, spike):
    return fig


def create_init_div(cfg_store):
    footprint_graph = dcc.Graph(figure=img_fig("cell candidates"))
    spike_graph = dcc.Graph(figure=spike_fig())

    status_div = html.Div(
        style=dict(
            display="grid",
            gridTemplateColumns="auto 100px 1fr",
        ),
        children=[
            button := dbc.Button("INIT"),
            session := html.Div(),
            progress := dbc.Progress(value="0"),
        ],
    )

    graph_div = html.Div(
        style=dict(
            display="grid",
            columnGap="10px",
            gridTemplateColumns="auto auto",
        ),
        children=[
            footprint_graph,
            spike_graph,
        ],
    )

    div = html.Div(
        style=dict(
            width="1200px",
        ),
        children=[
            status_div,
            graph_div,
        ],
    )

    @callback(
        Output(footprint_graph, "figure"),
        Input(button, "n_clicks"),
        State(cfg_store, "data"),
        background=True,
        interval=100,
        running=[
            (Output(button, "disabled"), True, False),
        ],
        progress=[
            Output(session, "children"),
            Output(progress, "value"),
            Output(progress, "max"),
            Output(progress, "label"),
        ],
        prevent_initial_call=True,
    )
    def get_init(set_progress, n_cilcks, cfg):
        pbar = Progress(set_progress)
        _cfg = OmegaConf.create(cfg)
        peaks = reduce(_cfg, pbar)
        footprint = make(_cfg, pbar)
        footprint_fig = update_img_fig(Patch(), footprint.max(axis=0), peaks.x, peaks.y)
        pbar.session("finish")
        return footprint_fig

    return div



from collections import namedtuple

from .regularizer import (
    Regularizer,
    L2,
    NonNegativeL1,
    MaxNormNonNegativeL1,
)

Penalty = namedtuple("Penalty", "la lu lb bs bt")


def get_penalty(penalty, stage=None):
    if isinstance(penalty, Penalty):
        return penalty
    out = []
    for p in (penalty.la, penalty.lu):
        match p:
            case {"type": "NoPenalty"}:
                out.append((Regularizer(),()))
            case {"type": "L2", "factor": fac}:
                out.append((L2, (fac,)))
            case {"type": "NonNegativeL1", "factor": fac}:
                out.append((NonNegativeL1(), (fac,)))
            case {"type": "MaxNormNonNegativeL1", "factor": fac}:
                out.append((MaxNormNonNegativeL1(), (fac,)))
            case {"type": "MaxNormNonNegativeL1", "annealing": fac, "max": maxval}:
                out.append((MaxNormNonNegativeL1(), (min(stage * fac, maxval),)))
            case _:
                raise ValueError()
    out += [penalty.lb, penalty.bs, penalty.bt]
    return Penalty(*out)



import sys
from pathlib import Path
from importlib import import_module
from multiprocessing import Process

from ..plot import (
    seg_max_fig,
    spike_image,
)
from .common import (
    finish,
    print_stats,
    set_env,
)


def call(name, *args, **kwargs):
    def wrap(cfg, *args, **kwargs):
        set_env(cfg)
        return target(cfg, *args, **kwargs)

    target = getattr(import_module(f"hotaru.cui.{name}"), name)
    p = Process(target=wrap, args=args, kwargs=kwargs)
    p.start()
    p.join()
    if p.exitcode != 0:
        sys.exit()


def run(cfg):
    for stage in range(cfg.max_train_step + 1):
        if stage == 0:
            call("normalize", cfg)
            call("init", cfg)
        else:
            call("spatial", cfg, stage)
        call("temporal", cfg, stage)

        print_stats(cfg, stage)
        if finish(cfg, stage):
            break

    path = Path(cfg.outputs.figs.dir)
    path.mkdir(parents=True, exist_ok=True)
    seg_max_fig(cfg, stage).write_image(path / "run_footprints.pdf")
    spike_image(cfg, stage)[0].save(path / "run_spike.pdf")



from functools import partial
from logging import getLogger

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf
from scipy.ndimage import grey_closing

from ..filter.laplace import gaussian_laplace_single
from ..utils import (
    from_tf,
    get_gpu_env,
)
from .segment import get_segment_mask

logger = getLogger(__name__)


def make_footprints(data, peaks, env=None, factor=1, prefetch=1):
    nk = peaks.shape[0]
    h, w = data.shape
    ts, ys, xs, rs = (np.array(v) for v in (peaks.t, peaks.y, peaks.x, peaks.radius))

    env = get_gpu_env(env)
    #nd = env.num_devices
    #sharding = env.sharding((nd, 1))
    batch = env.batch(float(factor) * h * w, nk)
    logger.info("make: %s batch=%d", peaks.shape[0], batch)

    logger.info("%s: %s %s %d", "pbar", "start", "make", ts.size)
    out = np.empty((nk, h, w))
    for r in np.unique(rs):
        index = np.where(rs == r)[0]
        dataset = tf.data.Dataset.from_generator(
            lambda: zip(index, data.data(ts[index]), ys[index], xs[index]),
            output_signature=(
                tf.TensorSpec((), tf.int32),
                tf.TensorSpec((h, w), tf.float32),
                tf.TensorSpec((), tf.int32),
                tf.TensorSpec((), tf.int32),
            ),
        )
        dataset = dataset.batch(batch)
        dataset = dataset.prefetch(prefetch)
        for d in dataset:
            d = (from_tf(v) for v in d)
            #idx, imgs, y, x = (jax.device_put(v, sharding) for v in d)
            idx, imgs, y, x = (jnp.array(v) for v in d)

            count = idx.size
            diff = batch - count
            if diff > 0:
                pad = (0, diff), (0, 0), (0, 0)
                imgs = jnp.pad(imgs, pad, constant_values=jnp.nan)
                y = jnp.pad(y, ((0, diff)), constant_values=-1)
                x = jnp.pad(x, ((0, diff)), constant_values=-1)
            out[idx] = np.array(_make_segments_simple(imgs, y, x, r)[:count])
            logger.info("%s: %s %d", "pbar", "update", count)
    out = grey_closing(out, (1, 10, 10))
    logger.info("%s: %s", "pbar", "close")
    return out


def make_segments_simple(imgs, y, x, r):
    return np.array(_make_segments_simple(imgs, y, x, r))


@partial(jax.jit, static_argnames=("r",))
def _make_segments_simple(imgs, y, x, r):
    g = gaussian_laplace_single(imgs, r)
    seg = jax.vmap(get_segment_mask)(g, y, x)
    val = jnp.where(seg, g, jnp.nan)
    dmin = jnp.nanmin(val, axis=(1, 2), keepdims=True)
    dmax = jnp.nanmax(val, axis=(1, 2), keepdims=True)
    out = jnp.where(seg, (val - dmin) / (dmax - dmin), 0)
    return out



import numpy as np
import plotly.graph_objects as go


def img_fig(title, colorscale="greens", scattercolor="red", zmin=None, zmax=None):
    heat = go.Heatmap(
        colorscale=colorscale,
        zmin=zmin,
        zmax=zmax,
        colorbar=dict(orientation="h", y=0, yanchor="top", thickness=10),
    )
    scatter = go.Scatter(mode="markers", marker_color=scattercolor)
    layout = go.Layout(
        title=title,
        xaxis=dict(visible=False, autorange=False),
        yaxis=dict(visible=False, autorange=False),
        width=600,
        height=570,
        margin=dict(t=25, b=45, l=0, r=0),
    )
    return go.Figure([heat, scatter], layout)


def update_img_fig(fig, img, x, y, height=500):
    h, w = img.shape
    fig.data[0].z = img
    fig.data[0].zmin = img.min()
    fig.data[0].zmax = img.max()
    fig.data[1].x = x
    fig.data[1].y = y
    fig.layout.width = height * w / h
    fig.layout.height = height + 70
    fig.layout.xaxis.range = [0, w]
    fig.layout.yaxis.range = [h, 0]
    return fig


def scatter_fig(title, xlabel, ylabel):
    scatter = go.Scatter(mode="markers", showlegend=False)
    layout = go.Layout(
        title=title,
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        width=600,
        height=570,
    )
    return go.Figure([scatter, scatter], layout)


def update_scatter_fig(fig, istd, icor, x, y):
    fig.data[0].x = istd[y, x]
    fig.data[0].y = icor[y, x]
    return fig


def radius_fig(title, xlabel, ylabel):
    fig = scatter_fig(title, xlabel, ylabel)
    fig.data[0].marker.maxdisplayed = 10000
    fig.data[0].marker.opacity = 0.3
    fig.data[1].marker.opacity = 0.3
    fig.layout.xaxis.type = "log"
    return fig


def update_radius_fig(fig, peakval, peaks):
    dr = peakval.radius[1] / peakval.radius[0] - 1
    v0 = peakval.v.ravel()
    r0 = np.array(peakval.radius)[peakval.r.ravel()]
    jitter0 = 1 + 0.2 * dr * np.random.randn(v0.size)
    jitter = 1 + 0.2 * dr * np.random.randn(peaks.r.size)
    fig.data[0].x = jitter0 * r0
    fig.data[0].y = v0
    fig.data[1].x = jitter * peaks.r
    fig.data[1].y = peaks.v
    return fig


def circle_fig(title):
    scatter = go.Scatter(mode="markers", marker=dict(color="green", sizeref=10.0))
    layout = go.Layout(
        title=title,
        xaxis=dict(visible=False, autorange=False),
        yaxis=dict(visible=False, autorange=False),
        width=600,
        height=570,
        margin=dict(t=25, b=45, l=0, r=0),
    )
    return go.Figure([scatter], layout)


def update_circle_fig(fig, peakval, peaks, height=500):
    h, w = peakval.v.shape
    fig.data[0].x = peaks.x
    fig.data[0].y = peaks.y
    fig.data[0].marker.opacity = 0.5 * peaks.v / peaks.v.max()
    fig.data[0].marker.size = peaks.r
    fig.data[0].marker.sizeref = 0.25 * h / 500
    fig.data[0].marker.sizemode = "diameter"
    fig.layout.width = height * w / h
    fig.layout.height = height + 70
    fig.layout.xaxis.range = [0, w]
    fig.layout.yaxis.range = [h, 0]
    return fig



import numpy as np
import plotly.graph_objects as go
import plotly.io as pio

from ..cui.common import load

pio.kaleido.scope.mathjax = None


def jitter(r, radius, scale):
    dscale = np.log((radius[1:] / radius[:-1]).min())
    jitter = np.exp(scale * dscale * np.random.randn(r.size))
    return r * jitter


def multi_peak_stats_fig(cfg, stage, stats, rmin, rmax):
    stats, _ = load(cfg, "evaluate", stage)
    fig = go.Figure().set_subplots(2, len(stats), row_heights=(1, 2))
    vmax = 0
    for i, peaks in enumerate(stats):
        cell = peaks.query("kind == 'cell'")
        if i == 0:
            intensity = "intensity"
        else:
            intensity = "firmness"
        v = cell[intensity]
        vmax = max(v.max(), vmax)
        fig.add_trace(
            go.Scattergl(
                x=cell.radius, #jitter(rs),
                y=v,
                mode="markers",
                marker=dict(opacity=0.3, size=5, color="green"),
                name="all pixels",
            ),
            col=i + 1,
            row=2,
        )
        r, c = np.unique(cell.radius, return_counts=True)
        print(r)
        print(c)
        fig.add_trace(
            go.Bar(
                x=np.log(r),
                y=c,
                marker_color="green",
            ),
            col=i + 1,
            row=1,
        )
        fig.update_xaxes(
            showticklabels=False,
            #tickmode="array",
            #tickvals=[np.log(2), np.log(4), np.log(8)],
            #ticktext=["2", "4", "8"],
            range=[np.log(rmin), np.log(rmax)],
            col=i + 1,
            row=1,
        )
        print(rmin, rmax)
        fig.update_xaxes(
            title_text="radius",
            type="log",
            tickmode="array",
            tickvals=[3, 6, 12],
            ticktext=["3", "6", "12"],
            autorange=False,
            range=[np.log10(rmin), np.log10(rmax)],
            col=i + 1,
            row=2,
        )
        if i == 0:
            fig.update_yaxes(
                title_text="intensity",
                col=1,
                row=2,
                range=(0, 1.05 * vmax),
            )
    fig.update_yaxes(
        title_text="fimness",
        col=2,
        row=2,
    )
    for i in range(1, len(stats)):
        fig.update_yaxes(
            range=(0, 1.05 * vmax),
            col=i + 1,
            row=2,
        )
    fig.update_yaxes(
        title_text="count",
        col=1,
        row=1,
    )
    return fig


def peak_stats_fig(cfg, stage, rmin=None, rmax=None, peakval=None):
    if stage == 0:
        stats = load(cfg, "init", stage)
    else:
        stats, _ = load(cfg, "evaluate", stage)
    if rmin is None:
        rmin = cfg.init.args.min_radius
    if rmax is None:
        rmax = cfg.init.args.max_radius
    peakval = None
    if stage == 0:
        try:
            peakval = load(cfg, "find", stage)
        except FileNotFoundError:
            pass

    fig = go.Figure().set_subplots(2, 1, row_heights=(1, 2))
    if peakval is None:
        intensity = "firmness"
        vmax = stats[intensity].max()
    else:
        intensity = "intensity"
        ri = peakval.r
        cond = ri > 0
        rs = peakval.radius[ri[cond]]
        vs = peakval.v[cond]
        vmax = vs.max()
        fig.add_trace(
            go.Scattergl(
                x=jitter(rs, peakval.radius, 0.3),
                y=vs,
                mode="markers",
                marker=dict(opacity=0.2, size=1, color="blue"),
                name="all pixels",
            ),
            col=1,
            row=2,
        )
        """
        cs = [(ri == i).sum() for i in np.arange(radius.size)]
        fig.add_trace(
            go.Bar(
                x=np.log(radius),
                y=cs,
            ),
            col=1,
            row=1,
        )
        """
    cell = stats.query("kind == 'cell'")
    fig.add_trace(
        go.Scattergl(
            x=cell.radius, #jitter(rs),
            y=cell[intensity],
            mode="markers",
            marker=dict(opacity=0.3, size=5, color="green"),
            name="all pixels",
        ),
        col=1,
        row=2,
    )
    r, c = np.unique(cell.radius, return_counts=True)
    fig.add_trace(
        go.Bar(
            x=np.log(r),
            y=c,
        ),
        col=1,
        row=1,
    )
    fig.update_xaxes(
        showticklabels=False,
        #tickmode="array",
        #tickvals=[np.log(2), np.log(4), np.log(8)],
        #ticktext=["2", "4", "8"],
        range=[np.log(rmin), np.log(rmax)],
        col=1,
        row=1,
    )
    fig.update_xaxes(
        title_text="radius",
        type="log",
        tickmode="array",
        tickvals=[3, 6, 12],
        ticktext=["3", "6", "12"],
        autorange=False,
        range=[np.log10(rmin), np.log10(rmax)],
        col=1,
        row=2,
    )
    fig.update_yaxes(
        title_text=intensity,
        range=(0, 1.05 * vmax),
        col=1,
        row=2,
    )
    fig.update_yaxes(
        title_text="count",
        col=1,
        row=1,
    )
    return fig


def peak_stats_trace(peaks, peakval=None, label=""):

    radius = np.sort(np.unique(peaks.radius))
    dscale = np.log((radius[1:] / radius[:-1]).min())
    data = []
    if peakval is None:
        intensity = "firmness"
        vmax = peaks[intensity].max()
    else:
        intensity = "intensity"
        ri = peakval.r
        cond = ri > 0
        rs = peakval.radius[ri[cond]]
        vs = peakval.v[cond]
        vmax = vs.max()
        data.append(
            go.Scattergl(
                x=jitter(rs),
                y=vs,
                mode="markers",
                marker=dict(opacity=0.01, color="blue"),
                name="all pixels",
            )
        )
    cell = peaks[peaks.kind == "cell"]
    bg = peaks[peaks.kind == "background"]
    data += [
        go.Scattergl(
            x=jitter(cell.radius),
            y=cell[intensity],
            mode="markers",
            marker=dict(size=10, symbol="star", color="green", opacity=0.1),
            name="cell",
        ),
        go.Scattergl(
            x=jitter(bg.radius),
            y=bg[intensity],
            mode="markers",
            marker=dict(size=10, symbol="pentagon", color="red", opacity=0.5),
            name="background",
        ),
    ]
    return data



from .clean import clean, clean_footprints
from .find import find_peaks
from .make import make_footprints
from .radius import get_radius
from .reduce import reduce_peaks, reduce_peaks_simple

__all__ = [
    "get_radius",
    "find_peaks",
    "reduce_peaks",
    "reduce_peaks_simple",
    "make_footprints",
    "clean",
    "clean_footprints",
]



import dash.dcc as dcc
import dash.html as html
import dash_bootstrap_components as dbc
from dash import (
    Input,
    Output,
    Patch,
    State,
    callback,
)
from omegaconf import OmegaConf

from ..main import (
    find,
    reduce,
)
from .graph import (
    circle_fig,
    radius_fig,
    update_circle_fig,
    update_radius_fig,
)
from .progress import Progress


def create_peaks_div(cfg_store):
    circle_graph = dcc.Graph(figure=circle_fig("cell candidates"))
    radius_graph = dcc.Graph(figure=radius_fig("", "radius", "indtensity"))

    status_div = html.Div(
        style=dict(
            display="grid",
            gridTemplateColumns="auto 100px 1fr",
        ),
        children=[
            button := dbc.Button("PEAKS"),
            session := html.Div(),
            progress := dbc.Progress(value="0"),
        ],
    )

    graph_div = html.Div(
        style=dict(
            display="grid",
            columnGap="10px",
            gridTemplateColumns="auto auto",
        ),
        children=[
            circle_graph,
            radius_graph,
        ],
    )

    div = html.Div(
        style=dict(
            width="1200px",
        ),
        children=[
            status_div,
            graph_div,
        ],
    )

    @callback(
        Output(circle_graph, "figure"),
        Output(radius_graph, "figure"),
        Input(button, "n_clicks"),
        State(cfg_store, "data"),
        background=True,
        interval=100,
        running=[
            (Output(button, "disabled"), True, False),
        ],
        progress=[
            Output(session, "children"),
            Output(progress, "value"),
            Output(progress, "max"),
            Output(progress, "label"),
        ],
        prevent_initial_call=True,
    )
    def get_peaks(set_progress, n_cilcks, cfg):
        pbar = Progress(set_progress)
        _cfg = OmegaConf.create(cfg)
        peakval = find(_cfg, pbar)
        peaks = reduce(_cfg, pbar)
        circle_fig = update_circle_fig(Patch(), peakval, peaks)
        radius_fig = update_radius_fig(Patch(), peakval, peaks)
        pbar.session("finish")
        return circle_fig, radius_fig

    return div



import jax.numpy as jnp


def get_radius(cfg):
    if isinstance(cfg, tuple):
        return cfg
    match cfg["type"]:
        case "logscale":
            r = jnp.geomspace(cfg["min"], cfg["max"], cfg["num"])
        case "linear":
            r = jnp.linspace(cfg["min"], cfg["max"], cfg["num"])
        case "list":
            r = jnp.array(cfg["val"])
    return tuple(float(ri) for ri in r)



import numpy as np
import pandas as pd
import h5py


def load(path, stage, df_only=False):
    if stage == 0:
        apath = f"{path}/../../../.."
        upath = f"{path}/../.."
    else:
        apath = path
        upath = path
    if df_only:
        return pd.read_csv(f"{upath}/{stage:03d}stats.csv", index_col=0)
    with h5py.File(f"{apath}/{stage:03d}footprints.h5") as h5:
        a = h5["data"][...]
    u = np.load(f"{upath}/{stage:03d}spike.npy")
    df = pd.read_csv(f"{upath}/{stage:03d}stats.csv", index_col=0)
    return a, u, df



import plotly.graph_objects as go
import numpy as np

from ..cui.common import load
from .common import add_jitter


def run_fig(
    cfg,
    stages,
    thr_f=0.35,
    thr_d=0.11,
    min_val=(0.25, 0),
    max_val=(0.58, 0.16),
    min_r=1,
    max_r=1,
):
    def get_color(x):
        if x.radius < 2.1:
            return "blue"
        elif (
            x.kind == "background"
            or x.radius > 15.9
            or x.old_udense > thr_d
            or x.firmness < thr_f
        ):
            return "red"
        else:
            return "green"

    fig = go.Figure().set_subplots(
        2 * len(stages),
        3,
        column_widths=(5, 5, 1),
        row_heights=(1, 5) * len(stages),
        horizontal_spacing=0.05,
        vertical_spacing=0.01,
        # shared_yaxes=True,
    )
    for i, stage in enumerate(stages):
        stats, _ = load(cfg, "evaluate", stage)
        df = stats.query("kind != 'remove'").copy()
        df = add_jitter(df)
        cl = df.query("kind == 'cell'")
        color = df.apply(get_color, axis=1)
        fig.add_trace(
            go.Histogram(x=df.firmness, marker_color="grey"),
            row=2 * i + 1,
            col=1,
        )
        fig.add_trace(
            go.Histogram(x=df.old_udense, marker_color="grey"),
            row=2 * i + 1,
            col=2,
        )
        fig.add_trace(
            go.Scatter(
                y=df.lri,
                x=df.firmness,
                mode="markers",
                marker=dict(size=3, color=color, opacity=0.5),
            ),
            row=2 * i + 2,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                y=df.lri,
                x=df.old_udense,
                mode="markers",
                marker=dict(size=3, color=color, opacity=0.5),
            ),
            row=2 * i + 2,
            col=2,
        )
        fig.add_trace(
            go.Histogram(y=cl.lri, marker_color="grey"),
            row=2 * i + 2,
            col=3,
        )
        fig.add_hline(y=(1 + min_r / 5), line_color="blue", row=2 * i + 2)
        fig.add_hline(y=(4 - max_r / 5), line_color="red", row=2 * i + 2)
        for r in range(2):
            fig.add_vline(thr_f, line_color="red", col=1, row=2 * i + r + 1)
            fig.add_vline(thr_d, line_color="red", col=2, row=2 * i + r + 1)
            for c in range(2):
                fig.update_xaxes(
                    range=(min_val[r], max_val[r]),
                    row=2 * i + 1,
                    col=1,
                )
                fig.update_xaxes(
                    range=(min_val[r], max_val[r]),
                    row=2 * i + 1,
                    col=2,
                )
        for c in range(3):
            fig.update_xaxes(
                showticklabels=False,
                row=2 * i + 1,
                col=1 + c,
            )
            fig.update_yaxes(
                showticklabels=False,
                row=2 * i + 1,
                col=1 + c,
            )
            fig.update_yaxes(
                title_text="radius" if c == 0 else "",
                title_font_size=11,
                showticklabels=c == 0,
                range=((4 / 5), (21 / 5)),
                tickmode="array",
                tickvals=np.log2([3, 6, 12]),
                ticktext=["3", "6", "12"],
                row=2 * i + 2,
                col=1 + c,
            )
            fig.update_xaxes(
                showticklabels=False,
                row=2 * i + 2,
                col=3,
            )
            fig.update_xaxes(
                showticklabels=False,
                range=(min_val[0], max_val[0]),
                row=2 * i + 1,
                col=1,
            )
            fig.update_xaxes(
                showticklabels=False,
                range=(min_val[1], max_val[1]),
                row=2 * i + 1,
                col=2,
            )
            bottom = i == len(stages) - 1
            fig.update_xaxes(
                showticklabels=bottom,
                range=(min_val[0], max_val[0]),
                title_text="firmness" if bottom else "",
                title_font_size=11,
                row=2 * i + 2,
                col=1,
            )
            fig.update_xaxes(
                showticklabels=bottom,
                range=(min_val[1], max_val[1]),
                title_text="spike density" if bottom else "",
                title_font_size=11,
                row=2 * i + 2,
                col=2,
            )
    fig.add_annotation(
        x=0.02,
        y=0.98,
        text="A",
        xref="paper",
        yref="paper",
        showarrow=False,
        font_size=13,
    )
    fig.add_annotation(
        x=0.02,
        y=0.48,
        text="B",
        xref="paper",
        yref="paper",
        showarrow=False,
        font_size=13,
    )
    fig.update_layout(
        template="none",
        font_size=11,
        width=600,
        height=200 * len(stages),
        showlegend=False,
        margin=dict(t=0, r=0, l=40, b=40),
    )
    return fig



from collections import namedtuple
from pathlib import Path
from logging import getLogger

import numpy as np
import pandas as pd
import h5py

logger = getLogger(__name__)


def h5_save(path, obj):
    layout = h5py.VirtualLayout(shape=obj.shape, dtype=obj.dtype)
    with h5py.File(path, "w") as h5:
        for i, val in enumerate(obj):
            my = np.where(np.any(val > 0, axis=1))[0]
            mx = np.where(np.any(val > 0, axis=0))[0]
            if my.size > 0:
                y0, y1, x0, x1 = my[0], my[-1] + 1, mx[0], mx[-1] + 1
                val = val[y0:y1, x0:x1]
                data = h5.create_dataset(f"{i}", data=val)
                layout[i, y0:y1, x0:x1] = h5py.VirtualSource(data)
        h5.create_virtual_dataset("data", layout, fillvalue=0)


def h5_load(path):
    with h5py.File(path, "r") as h5:
        return h5["data"][...]


def save(path, obj):
    if isinstance(path, (list, tuple)):
        if len(path) == 1:
            save(path[0], obj[0])
        else:
            for p, o in zip(path, obj):
                save(p, o)
    else:
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        match path.suffix:
            case ".npz":
                np.savez(path, **obj._asdict())
            case ".npy":
                np.save(path, obj)
            case ".csv":
                obj.to_csv(path)
            case ".h5":
                h5_save(path, obj)
            case ".flag":
                path.write_text(obj)
            case _:
                raise ValueError()


def try_load(path):
    if isinstance(path, (list, tuple)):
        if len(path) == 1:
            return try_load(path[0])
        else:
            return [try_load(p) for p in path]
    else:
        path = Path(path)
        logger.debug("%s", path)
        try:
            match path.suffix:
                case ".npz":
                    with np.load(path) as npz:
                        return namedtuple("LoadedData", npz.files)(**dict(npz.items()))
                case ".npy":
                    return np.load(path)
                case ".csv":
                    return pd.read_csv(path, index_col=0)
                case ".h5":
                    return h5_load(path)
                case ".flag":
                    return path.read_text()
                case _:
                    return
        except Exception as e:
            logger.debug("%s", e)
            return



import pandas as pd

from hotaru.plot.compla import compare_la_fig


paths = [
    "outputs/CA3/miniature/min21/default/u100/a0/dup07-30-8/001stats.csv",
    "outputs/CA3/miniature/min21/default/u100/default/dup07-30-8/001stats.csv",
]
compare_la_fig(paths, ["a0", "a20"]).write_image(
    "figs/la.pdf",
)



import numpy as np
import pandas as pd
import plotly.express as px
import plotly.io as pio

pio.kaleido.scope.mathjax = None


def plot_stats(paths, imgs, labels, dlabels, **kwargs):
    dfs = []
    for dlabel, path in zip(dlabels, paths):
        df = pd.DataFrame()
        for label, img in zip(labels, imgs):
            img = np.load(f"{path}/{img}.npy")
            df[label] = img.ravel()
        df["Data"] = dlabel
        dfs.append(df)
    df = pd.concat(dfs, axis=0)
    fig = px.scatter(
        df, x=labels[1], y=labels[2], color=labels[0], facet_col="Data", **kwargs
    )
    fig.update_xaxes(
        title_font_size=11,
    )
    fig.update_yaxes(
        title_font_size=11,
    )
    fig.update_layout(
        coloraxis_colorbar_title_font_size=11,
        template="none",
        font_size=11,
        margin=dict(l=50, r=20, t=20, b=50),
        annotations=[
            dict(font=dict(size=11), text=dlabels[i])
            for i in range(len(fig.layout.annotations))
        ],
    )
    return fig



import numpy as np
import plotly.express as px

from ..cui.common import all_stats


def cell_num_fig(cfg, **kwargs):
    kwargs.setdefault("template", "none")
    kwargs.setdefault("margin", dict(l=40, r=10, t=20, b=35))

    num = [np.count_nonzero(s.kind == "cell") for s in all_stats(cfg)]

    fig = px.line(x=np.arange(len(num)), y=num)
    fig.add_annotation(x=1, y=num[1], ax=2, text="A")
    fig.add_annotation(x=len(num)-1, y=num[-1], text="B")
    fig.update_xaxes(
        title_text="epoch",
        title_font_size=11,
    )
    fig.update_yaxes(
        title_text="Number of cells",
        title_font_size=11,
    )
    fig.update_layout(**kwargs)
    return fig



import jax.lax as lax
import jax.numpy as jnp


def max_pool(inputs, window, strides, padding):
    ndiff = inputs.ndim - len(window)
    window = (1,) * ndiff + window
    strides = (1,) * ndiff + strides
    return lax.reduce_window(inputs, -jnp.inf, lax.max, window, strides, padding)



from logging import getLogger

import jax
import jax.lax as lax
import jax.numpy as jnp
import numpy as np

logger = getLogger(__name__)


class ProxOptimizer:
    def __init__(self, model):
        self._model = model
        self._loss_sep_fn = jax.jit(self._loss_sep)
        self._loss_fn = jax.jit(self._loss)
        self._step_fn = jax.jit(self._step, static_argnames=("n_step",))

    def loss(self, *x):
        loss_args = dict(
            args=tuple(jnp.array(v) for v in self._model.args),
            loss_scale=jnp.array(self._model.loss_scale),
            prox_args=tuple(
                tuple(jnp.array(vi) for vi in v) for v in self._model.prox_args
            ),
        )
        x = tuple(jnp.array(xi) for xi in x)
        loss, aux = self._loss_fn(*x, **loss_args)
        return np.array(loss), np.array(aux)

    def step(self, x, lr=1, nesterov=20, n_step=100):
        x = tuple(jnp.array(xi) for xi in x)
        args = tuple(jnp.array(v) for v in self._model.args)
        prox_args = tuple(tuple(jnp.array(vi) for vi in v) for v in self._model.args)
        lr = jnp.array(lr, jnp.float32)
        nesterov = jnp.array(nesterov, jnp.float32)
        return self._step_fn(x, args, prox_args, lr, nesterov, n_step)

    def fit(self, x, max_epoch, steps_par_epoch, lr, nesterov, tol, patience, env):
        common_args = dict(
            args=tuple(jnp.array(v) for v in self._model.args),
            prox_args=tuple(
                tuple(jnp.array(vi) for vi in v) for v in self._model.prox_args
            ),
        )
        loss_args = common_args | dict(loss_scale=jnp.array(self._model.loss_scale))
        step_args = common_args | dict(
            lr=jnp.array(lr, jnp.float32),
            nesterov=jnp.array(nesterov, jnp.float32),
            n_step=steps_par_epoch,
        )

        def loss_fn(*x):
            loss, aux = self._loss_fn(*x, **loss_args)
            return np.array(loss), np.array(aux)

        x = tuple(jnp.array(xi) for xi in x)

        history = []

        loss, aux = loss_fn(*x)
        diff = np.nan
        history.append((loss, aux))

        log_diff = np.inf
        postfix = f"loss={loss:.4f}, diff={log_diff:.2f}"
        logger.info("%s: %s %d %s", "pbar", "update", 0, postfix)

        patience_count = 0
        min_loss = np.inf
        for i in range(max_epoch):
            x = self._step_fn(x, **step_args)

            (loss, aux), old_loss = loss_fn(*x), loss
            diff = (old_loss - loss) / tol
            history.append((loss, aux))

            log_diff = np.log10(diff) if diff > 0 else np.nan
            postfix = f"loss={loss:.4f}, diff={log_diff:.2f}, aux={aux:.4f}"
            logger.info("%s: %s %d %s", "pbar", "update", 1, postfix)

            if loss > min_loss - tol:
                patience_count += 1
            else:
                patience_count = 0
            if patience_count == patience:
                break
            min_loss = min(loss, min_loss)
        return tuple(np.array(xi) for xi in x), history

    def _loss_sep(self, *x, args, loss_scale, prox_args):
        loss, aux = self._model.loss(*x, *args)
        penalty = sum(
            ri(xi, *ai) for xi, ri, ai in zip(x, self._model.regularizers, prox_args)
        )
        return loss / loss_scale, penalty / loss_scale, aux

    def _loss(self, *x, **args):
        loss, penalty, aux = self._loss_sep(*x, **args)
        return loss + penalty, aux

    def _step(self, x, args, prox_args, lr, nesterov, n_step):
        def update(i, xy):
            oldx, oldy = xy
            grady, aux = grad_loss_fn(*oldy, *args)
            t0 = (nesterov + i) / (nesterov)
            t1 = (nesterov + 1) / (nesterov + i + 1)
            newx = tuple(
                p(y - lr * g, lr, *a)
                for p, a, y, g in zip(prox, prox_args, oldy, grady)
            )
            tmpx = tuple((1 - t0) * xo + t0 * xn for xo, xn in zip(oldx, newx))
            newy = tuple((1 - t1) * xn + t1 * xt for xn, xt in zip(newx, tmpx))
            return newx, newy

        grad_loss_fn = jax.grad(self._model.loss, range(len(x)), has_aux=True)
        prox = self._model.prox
        x, _ = lax.fori_loop(0, n_step, update, (x, x))
        return x



import numpy as np


def calc_distance(x, y, r):
    return np.hypot(x - x[:, np.newaxis], y - y[:, np.newaxis]) / r


def cross_image(cfg, stage, mx=None, hsize=30, pad=5)
    def s(i):
        return pad + i * (size + pad)

    def e(i):
        return (i + 1) * (size + pad)

    def _clip(j, k)
        return slice(s(j), e(j)), slice(s(k), e(k))

    size = 2 * hsize + 1

    if stage == 0:
        segs = load(cfg, "make", stage)
        stats = load(cfg, "init", stage)
    else:
        segs, stats = load(cfg, "clean", stage)
    cell = stats.query("kind == 'cell'").index
    state = state[cell]
    fp = segs[cell]

    nk = stats.shape[0]
    ys = stats.y.to_numpy()
    xs = stats.x.to_numpy()
    rs = stats.radius.to_numpy()

    d = calc_distance(x, y, r)
    k0, k1 = np.where(dist < 3)
    nk = k0.shape

    if mx is None:
        mx = int(np.floor(np.sqrt(nk)))
    my = (nk + mx - 1) // mx

    segs = np.pad(segs, ((0, 0), (hsize, hsize), (hsize, hsize)))
    clip0 = np.zeros(
        (my * size + pad * (my + 1), mx * size + pad * (mx + 1), 4), np.uint8
    )
    clip1 = np.zeros(
        (my * size + pad * (my + 1), mx * size + pad * (mx + 1), 4), np.uint8
    )

    for i range(nk):
        j, k = divmod(i, mx)
        k0, k1 = k0s[i], k1s[i]
        y0, x0, y1, x1 = ys[k0], xs[k0], ys[k1], xs[k1]
        c = _clip(j, k)
        clip0[c] = to_image(
            fp[k0, y0 : y0 + size, x0 : x0 + size],
            "Blues",
        )
        clip1[c] = to_image(
            fp[k1, y1:y1+size, x1: x1 + size],
            "Reds",
        )

    for x in range(mx + 1):
        st = x * (size + pad)
        en = st + pad
        clip[:, st:en] = [0, 0, 0, 255]
    for y in range(my + 1):
        st = y * (size + pad)
        en = st + pad
        clip[st:en] = [0, 0, 0, 255]

    return Image.fromarray(clip)


def cross_fig(cfg, stage, hsize=30, **kwargs):
    width = kwargs.setdefault("width", 600)

    segs, _ = load(cfg, "clean", stage)
    spk, _ = load(cfg, "temporal", stage)
    state, _ = load(cfg, "evaluate", stage)
    cell = state.query("kind == 'cell'").index
    fp = segs[cell]
    skp = spk[cell]
    spk /= spk.max(axis=1, keepdims=True)

    d = calc_distance(state)
    k0, k1 = np.where(dist < 3)

    fp= np.pad(fp, ((0, 0), (hsize, hsize), (hsize, hsize)))

    nk = k0.size
    kwargs.setdefault("height", width / 10 * nk)

    fig = go.Figure().set_subplots(2, nk, column_width=(1, 9)
    for i, k0i, k1i in enumerate(zip(k0[:10], k1[:10])):
        print(k0i, k1i, d[k0i, k1i])
        fig = go.Figure()
        y0, x0 = y[k0i], x[k0i]
        y1, x1 = y[k1i], x[k1i]
        ym, xm = (y0 + y1) // 2, (x0 + x1) // 2
        fig.add_trace(
            go.Heatmap(
                z=fp[k0i, ym : ym + size, xm : xm + size],
                opacity=0.8, colorscale="Blues", zmin=0, zmax=1, showscale=False,
            ),
            row=i + 1,
            col=1,
        )
        fig.add_trace(
            go.Heatmap(
                z=fp[k1i, ym : ym + size, xm : xm + size],
                opacity=0.8, colorscale="Blues", zmin=0, zmax=1, showscale=False,
            ),
            row=i + 1,
            col=1,
        )
        fig.update_xaxes(
            showticklabels=False,
            row=i + 1,
            col=1,
        )
        fig.update_yaxes(
            showticklabels=False,
            row=i + 1,
            col=1,
        )
        fig.add_trace(
            go.Heatmap(
                z=spk[[k0i, k1i]],
                colorscale="Reds", zmin=0, zmax=1, showscale=False,
            ),
            row=i + 1,
            col=2,
        )
    fig.update_layout(
    )



from threading import Thread

import dash.dcc as dcc
import dash.html as html
import dash_bootstrap_components as dbc
from dash import (
    Input,
    Output,
    State,
    callback,
    ctx,
    no_update,
)
from omegaconf import OmegaConf


class ProgressProxy:
    def __init__(self):
        self()

    def __call__(self, total=1, desc=None, postfix=None):
        self.desc = desc or ""
        self.postfix = postfix or ""
        self.total = total
        self.n = 0
        return self

    def set_description(self, desc):
        self.desc = desc

    def set_postfix_str(self, postfix):
        self.postfix

    def reset(self, total):
        self.total = total
        self.n = 0

    def update(self, n):
        self.n += n

    @property
    def value(self):
        return self.n / self.total

    @property
    def label(self):
        return f"{self.desc}: {int(100 * self.n / self.total)}% ({self.postfix})"


def two_column(width):
    return dict(
        width=f"{width}px",
        display="grid",
        gridTemplateColumns="auto 1fr",
    )


def ThreadButton(label, func, cfg, pbar):
    div = html.Div(
        children=[
            finish := dcc.Store(label),
            button := dbc.Button(label),
            interval := dcc.Interval(interval=100, disabled=True),
        ],
        style=two_column(1200),
    )
    jobs = [None, None]

    @callback(
        Output(finish, "data"),
        Output(interval, "disabled"),
        Output(pbar, "value"),
        Output(pbar, "label"),
        Input(button, "n_clicks"),
        Input(interval, "n_intervals"),
        State(cfg, "data"),
        prevent_initial_call=True,
    )
    def on_click(n_clicks, n_intervals, cfg):
        if ctx.triggered_id == button.id:
            cfg = OmegaConf.create(cfg)
            pbar = ProgressProxy()
            thread = Thread(target=func, kwargs=dict(cfg=cfg, pbar=pbar))
            jobs[:] = thread, pbar
            thread.start()
            return no_update, False, 0, pbar.label
        elif ctx.triggered_id == interval.id:
            thread, pbar = jobs
            if thread.is_alive():
                return no_update, no_update, pbar.value, pbar.label
            else:
                return "finish", True, 100, pbar.label

    div.finish = Input(finish, "data")
    return div


def Collapse(name, is_open, *children):
    div = html.Div(
        children=[
            button := dbc.Button(f"Open/Close {name}"),
            collapse := dbc.Collapse(children=children, is_open=is_open),
        ]
    )

    @callback(
        Output(collapse, "is_open"),
        Input(button, "n_clicks"),
        State(collapse, "is_open"),
        prevent_initial_call=True,
    )
    def toggle_stats(n, is_open):
        return not is_open

    return div



import pandas as pd
import plotly.express as px
import plotly.io as pio

from ..cui.common import load

pio.kaleido.scope.mathjax = None


def dense_sig_fig(cfg, stages, thr_sig=0, label="", thr_udense=1.0):
    num = []
    dfs = []
    for stage in stages:
        df, _ = load(cfg, "evaluate", stage)
        df = df.query("kind != 'remove'")
        df.loc[df.udense > thr_udense, "kind"] = "background"
        df["col"] = stage
        df["size"] = 1
        df["symbol"] = 1
        dfs.append(df)
        num.append(df.query("kind == 'cell'").shape[0])
    df = pd.concat(dfs, axis=0)
    fig = px.scatter(
        df,
        x="udense",
        y="signal",
        color="kind",
        size="size",
        size_max=3,
        opacity=0.5,
        facet_col="col",
    )
    for i in range(len(fig.data)):
        fig.data[i].marker.line.width = 0

    for n, stage in zip(num, stages):
        fig.update_annotations(
            selector=dict(text=f"col={stage}"), y=0.8, text=f"{label}{stage}; num={n}"
        )
    fig.update_layout(
        template="none",
        font_size=11,
        showlegend=False,
        width=600,
        height=150,
        margin=dict(l=35, r=10, t=20, b=32),
    )
    return fig


def dense_sig_multi_fig(paths, stages, thr_sig=0, thr_udense=1.0):
    num = []
    dfs = []
    for i, (name, path) in enumerate(paths.items()):
        for j, stage in enumerate(stages[i]):
            if stage == 0:
                pathx = f"{path}/../../{stage:03d}stats.csv"
            else:
                pathx = f"{path}/{stage:03d}stats.csv"
            df = pd.read_csv(pathx)
            df = df.query("kind != 'remove'")
            df.loc[df.udense > thr_udense, "kind"] = "background"
            df["name"] = name
            df["stage"] = j
            df["size"] = 1
            df["symbol"] = 1
            dfs.append(df)
            num.append(df.query("kind == 'cell'").shape[0])
    df = pd.concat(dfs, axis=0)
    fig = px.scatter(
        df,
        x="udense",
        y="signal",
        color="kind",
        size="size",
        size_max=3,
        opacity=0.5,
        facet_col="stage",
        facet_row="name",
    )
    for i in range(len(fig.data)):
        fig.data[i].marker.line.width = 0
    for i, name in enumerate(paths.keys()):
        fig.update_annotation(
            selector=dict(text=f"naem={name}"),
            text=name,
        )
    for j, stage in enumerate(paths.keys()):
        fig.update_annotation(
            selector=dict(text=f"naem={name}"),
            text="",
        )
    fig.update_layout(
        template="none",
        font_size=11,
        showlegend=False,
        width=600,
        height=300,
        margin=dict(l=35, r=10, t=20, b=32),
    )
    return fig



import multiprocessing as mp
from logging import getLogger

import numpy as np
import pandas as pd

logger = getLogger(__name__)


def reduce_peaks_simple(
    ys, xs, rs, vs, min_radius, max_radius, min_distance_ratio, old_bg=None
):
    if old_bg is None:
        old_bg = []

    n = np.count_nonzero(np.isfinite(vs))
    flg = np.argsort(vs)[::-1][:n]
    cell = []
    bg = []
    remove = []
    while flg.size > 0:
        i, flg = flg[0], flg[1:]
        r0 = rs[i]
        if r0 >= min_radius:
            y0, x0 = ys[i], xs[i]
            if i in old_bg or r0 > max_radius:
                yb, xb = ys[bg], xs[bg]
                distance = np.hypot(xb - x0, yb - y0) / r0
                if not bg or np.all(distance >= min_distance_ratio):
                    bg.append(i)
                    logger.debug(
                        "background: id=%d old=%d r=%f dist=%s",
                        i,
                        i in old_bg,
                        r0,
                        sorted(distance)[:2],
                    )
                else:
                    remove.append(i)
                    logger.debug(
                        "removed (dup bg): id=%d old=%d r=%f dist=%s",
                        i,
                        i in old_bg,
                        r0,
                        sorted(distance)[:2],
                    )
            else:
                yc, xc = ys[cell], xs[cell]
                distance = np.hypot(xc - x0, yc - y0) / r0
                if not cell or np.all(distance >= min_distance_ratio):
                    y1, x1 = ys[flg], xs[flg]
                    dist2 = np.hypot(x1 - x0, y1 - y0) / r0
                    cond = dist2 > min_distance_ratio
                    if np.any(~cond):
                        remove += list(flg[~cond])
                        logger.debug(
                            "pre remove: %s dist=%s", list(flg[~cond]), dist2[~cond]
                        )
                    flg = flg[cond]
                    cell.append(i)
                    logger.debug("cell: id=%d dist%s", i, sorted(distance)[:2])
                else:
                    remove.append(i)
                    logger.debug(
                        "removed (dup cell): id=%d r=%f dist=%s",
                        i,
                        r0,
                        sorted(distance)[:2],
                    )
        else:
            remove.append(i)
            logger.debug("removed (small r): id=%d r=%f", i, r0)
    return cell, bg, remove


def reduce_peaks_mesh(rs, vs, *args, **kwargs):
    h, w = rs.shape
    ys, xs = np.mgrid[:h, :w]
    ys, xs, rs, vs = ys.ravel(), xs.ravel(), rs.ravel(), vs.ravel()
    cell, bg, remove = reduce_peaks_simple(ys, xs, rs, vs, *args, **kwargs)
    return ys[cell], xs[cell], ys[bg], xs[bg]


def reduce_peaks(
    peakval,
    bg_type="bg",
    min_radius=None,
    max_radius=None,
    min_distance_ratio=None,
    block_size=None,
):
    logger.info(
        "reduce_peaks: %f %f %f %d",
        min_radius,
        max_radius,
        min_distance_ratio,
        block_size,
    )
    logger.info("bg_type %s", bg_type)
    static_args = min_radius, max_radius, min_distance_ratio

    radius, ts, ri, vs = peakval
    rs = radius[ri]
    h, w = rs.shape
    margin = int(np.ceil(min_distance_ratio * rs.max()))

    args = []
    for xs in range(0, w - margin, block_size):
        x0 = max(xs - margin, 0)
        xe = xs + block_size
        x1 = min(xe + margin, w)
        for ys in range(0, h - margin, block_size):
            y0 = max(ys - margin, 0)
            ye = ys + block_size
            y1 = min(ye + margin, h)
            r = rs[y0:y1, x0:x1]
            v = vs[y0:y1, x0:x1]
            args.append(((y0, x0, ys, xs, ye, xe), (r, v, *static_args)))

    out = []
    with mp.Pool() as pool:
        for o in pool.imap_unordered(_reduce_peaks, args):
            out.append(o)
    celly, cellx, bgy, bgx = [np.concatenate(v, axis=0) for v in zip(*out)]

    def make_dataframe(y, x):
        df = pd.DataFrame(
            dict(
                y=y,
                x=x,
                t=ts[y, x],
                radius=rs[y, x],
                firmness=vs[y, x],
            )
        )
        return df.sort_values("firmness", ascending=False)

    cell = make_dataframe(celly, cellx)
    cell["kind"] = "cell"
    if bg_type == "remove":
        peaks = cell
    elif bg_type == "bg":
        bg = make_dataframe(bgy, bgx)
        bg["kind"] = "background"
        peaks = pd.concat([cell, bg], axis=0)
    else:
        raise ValueError()
    peaks = peaks.reset_index(drop=True)
    peaks.insert(0, "segid", peaks.index)
    nk = np.count_nonzero(peaks.kind == "cell")
    nb = np.count_nonzero(peaks.kind == "background")
    logger.info("num cell/bg: %d/%d", nk, nb)
    return peaks[["segid", "kind", "y", "x", "t", "radius", "firmness"]]


def _reduce_peaks(args):
    def fix(y, x):
        y += y0
        x += x0
        cond = (ys <= y) & (y < ye) & (xs <= x) & (x < xe)
        return y[cond], x[cond]

    y0, x0, ys, xs, ye, xe = args[0]
    celly, cellx, bgy, bgx = reduce_peaks_mesh(*args[1])
    return *fix(celly, cellx), *fix(bgy, bgx)



import webbrowser
from pathlib import Path
from importlib.resources import read_text

import hydra

from .cui import (
    run,
    test,
)


@hydra.main(version_base=None, config_path="pkg://hotaru.conf", config_name="config")
def _main(cfg):
    match cfg.mode:
        case "test":
            test(cfg)
        case "run":
            run(cfg)
        case _:
            raise ValueError()


def main():
    cfg_path = Path("hotaru/default.yaml")
    if not cfg_path.exists():
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        cfg = read_text("hotaru.conf", "user_sample.yaml")
        cfg_path.write_text(cfg)

    _main()


if __name__ == "__main__":
    main()



from collections import namedtuple
from functools import wraps
from pathlib import Path

import numpy as np
import pandas as pd

from ..filter.stats import calc_stats
from ..footprint.clean import clean_segment_batch
from ..footprint.find import find_peaks_batch
from ..footprint.make import make_segment_batch
from ..footprint.reduce import reduce_peaks_block
from ..io.image import load_imgs
from ..io.saver import (
    load,
    save,
)
from ..proxmodel.regularizer import MaxNormNonNegativeL1
from ..train.dynamics import SpikeToCalcium
from ..train.loss import (
    gen_factor,
    gen_optimizer,
)
from .progress import (
    SilentProgress,
    TQDMProgress,
)

Data = namedtuple("Data", "imgs mask avgx avgt std0")
Stats = namedtuple("Stats", "imin imax istd icor")
Penalty = namedtuple("Penalty", "la lu bx bt")


def HagetakaCUI(cfg):
    pbar = TQDMProgress()
    nk = -1
    for stage in range(cfg.total_stage + 1):
        old_nk, nk = nk, temporal_optimize(cfg, stage, pbar).shape[0]
        if old_nk == nk:
            break


def autoload(func, cfg, stage=None, pbar=None):
    @wraps(func)
    def wrap(cfg, stage=None, pbar=None):
        if pbar is None:
            pbar = SilentProgress()
        cfg.stage, stage_saved = stage, cfg.stage
        c = cfg.solver
        for k in func.__name__.split("_"):
            c = c.get(k)
        file = Path(cfg.outdir) / c.outfile[0]
        force = c.force
        if stage is not None:
            force = force or (c.done is not None)
            if force:
                if c.done is None:
                    c.done = -1
                force = stage > c.done
        if not force and file.exists():
            obj = load(file)
            pbar.skip()
        else:
            save_objs = func(cfg, c, stage, pbar)
            cfg.stage = stage_saved
            if not isinstance(save_objs, tuple):
                save_objs = (save_objs,)
            for obj in save_objs:
                save(file, obj)
            if stage is None:
                c.force = False
            else:
                c.done = stage
        cfg.stage = stage_saved
        return save_objs
    return wrap


def radius(c):
    return np.geomspace(c.rmin, c.rmax, c.rnum)


def dynamics(cfg):
    c = cfg.dynamics
    if c.kind == "double_exp":
        return SpikeToCalcium.double_exp(c.tau1, c.tau2, c.duration, cfg.data.hz)


def penalty(cfg, c):
    imgs, mask = load_imgs(cfg.data)
    if mask is None:
        nt, h, w = imgs.shape
        scale = nt * h * w
    else:
        scale = imgs.shape[0] * np.count_nonzero(mask)
    la = MaxNormNonNegativeL1(c.la / scale)
    lu = MaxNormNonNegativeL1(c.lu / scale)
    return Penalty(la, lu, c.bx, c.bt)


@autoload
def stats(cfg, c, stage, pbar):
    imgs, mask = load_imgs(cfg.data)
    return calc_stats(imgs, mask, c.batch, pbar.session("stats"))


def imgs_stats(cfg, pbar=None):
    s = stats(cfg, pbar=pbar)
    return Stats(s.imin, s.imax, s.istd, s.icor)


def get_frame(cfg, t, pbar=None):
    imgs, mask = load_imgs(cfg.data)
    s = stats(cfg, pbar=pbar)
    return (imgs[t] - s.avgx - s.avgt[t]) / s.std0


def data(cfg, pbar=None):
    imgs, mask = load_imgs(cfg.data)
    s = stats(cfg, pbar=pbar)
    return Data(imgs, mask, s.avgx, s.avgt, s.std0)


@autoload
def find(cfg, c, stage, pbar):
    return find_peaks_batch(
        data(cfg, pbar=pbar),
        radius(c),
        c.batch,
        pbar.session("find"),
    )


@autoload
def reduce(cfg, c, stage, pbar):
    return reduce_peaks_block(
        find(cfg, pbar=pbar),
        c.rmin,
        c.rmax,
        c.thr,
        c.block_size,
    )


@autoload
def make(cfg, c, stage, pbar):
    return make_segment_batch(
        data(cfg, pbar=pbar),
        reduce(cfg, pbar=pbar),
        c.batch,
        pbar.session("make"),
    )


@autoload
def clean(cfg, c, stage, pbar):
    imgs, mask = load_imgs(cfg.data)
    val = spatial_optimize(cfg, stage, pbar)
    val /= val.max(axis=1, keepdims=True)
    nk, h, w = val.shape[0], *imgs.shape[1:]
    if mask is None:
        fimgs = val.reshape(nk, h, w)
    else:
        fimgs = np.empty((nk, h, w), np.float32)
        fimgs[:, mask] = val
    cimgs, peaks = clean_segment_batch(
        fimgs,
        mask,
        radius(c),
        c.batch,
        pbar.session(f"{stage}footprint clean"),
    )
    cimgs = cimgs[peaks.r > 0]
    return cimgs, pd.DataFrame(peaks._asdict())


def temporal_prepare(cfg, c, stage, pbar):
    return gen_factor(
        "temporal",
        make(cfg, pbar=pbar) if stage == 0 else clean(cfg, stage, pbar)[0],
        data(cfg, pbar=pbar),
        dynamics(cfg),
        penalty(cfg, cfg.temporal.penalty),
        c.batch,
        pbar.session(f"{stage}spike prepare"),
    )


@autoload
def temporal_optimize(cfg, c, stage, pbar):
    optimizer = gen_optimizer(
        "temporal",
        temporal_prepare(cfg, stage, pbar),
        dynamics(cfg),
        penalty(cfg, cfg.temporal.penalty),
        c.lr,
        c.scale,
        c.loss.num_devices,
    )
    optimizer.fit(
        c.n_epoch,
        c.n_step,
        c.early_stop.tol,
        pbar.session(f"{stage}spike optimize"),
        c.prox.num_devices,
    )
    return optimizer.val[0]


@autoload
def spatial_prepare(cfg, c, stage, pbar):
    spk = temporal_optimize(cfg, stage - 1, pbar)
    spk /= spk.max(axis=1, keepdims=True)
    return gen_factor(
        "spatial",
        spk,
        data(cfg, pbar=pbar),
        dynamics(cfg),
        penalty(cfg, cfg.penalty),
        c.batch,
        pbar.session(f"{stage}footprint prepare"),
    )


@autoload
def spatial_optimize(cfg, c, stage, pbar):
    optimizer = gen_optimizer(
        "spatial",
        spatial_prepare(cfg, stage, pbar),
        dynamics(cfg),
        penalty(cfg, cfg.penalty),
        c.lr,
        c.scale,
        c.loss.num_devices,
    )
    optimizer.fit(
        c.n_epoch,
        c.n_step,
        c.early_stop.tol,
        pbar.session(f"{stage}footprint optimize"),
        c.prox.num_devices,
    )
    return optimizer.val[0]



import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio

pio.kaleido.scope.mathjax = None

# from hotaru.plot.densesig import dense_sig_multi_fig


paths = dict(
    u0="outputs/Kd32/miniature/min31/clip2x2/u0/default/dup9/",
    u40="outputs/Kd32/miniature/min31/clip2x2/u40/default/dup9/",
)
stages = [[0, 1, 17], [0, 1, 14]]
thr_udense = 0.18

xmax = 0.5
ymax = 4.5

fig = go.Figure().set_subplots(2, 3, vertical_spacing=0.03, horizontal_spacing=0.02)
for i, (name, path) in enumerate(paths.items()):
    fig.add_annotation(
        text=name,
        x=0.15,
        y=4,
        xanchor="left",
        showarrow=False,
        col=1,
        row=1 + i,
    )
    for j, stage in enumerate(stages[i]):
        if stage == 0:
            pathx = f"{path}/../../{stage:03d}stats.csv"
        else:
            pathx = f"{path}/{stage:03d}stats.csv"
        df = pd.read_csv(pathx)
        df = df.query("kind != 'remove'")
        if j == 2:
            df.loc[df.udense > thr_udense, "kind"] = "background"
        fig.add_trace(
            go.Scatter(
                x=df.udense,
                y=df.signal,
                mode="markers",
                marker=dict(
                    color=[dict(cell="blue", background="red")[k] for k in df.kind],
                    size=3,
                    opacity=0.2,
                    line_width=0,
                )
            ),
            col=1 + j,
            row=1 + i,
        )
        fig.update_xaxes(
            title="spike density" if i == 1 else "",
            showticklabels=i == 1,
            range=(0, xmax),
            col=1 + j,
            row=1 + i,
        )
        fig.update_yaxes(
            title="signal" if j == 0 else "",
            showticklabels=j == 0,
            range=(0, ymax),
            col=1 + j,
            row=1 + i,
        )
        num = df.query("kind == 'cell'").shape[0]
        fig.add_annotation(
            text=f"epoch={stages[i][j]}, num={num}",
            x=0.2,
            y=3,
            xanchor="left",
            showarrow=False,
            col=1 + j,
            row=1 + i,
        )
fig.update_layout(
    template="none",
    showlegend=False,
    margin=dict(l=50, t=10, b=50, r=10),
    width=600,
    height=300,
    font_size=11,
)

fig.write_image("figs/Kd32-dense-sig.pdf")



from .data import Data
from .clip import get_clip
from .timer import Timer
from .gpu import (
    from_tf,
    get_gpu_env,
    get_gpu_info,
    get_gpu_used,
    get_xla_stats,
    delete_xla_buffers,
)
from .progress import get_progress

__all__ = [
    "Data",
    "Timer",
    "get_clip",
    "get_progress",
    "from_tf",
    "get_gpu_env",
    "get_gpu_info",
    "get_gpu_used",
    "get_xla_stats",
    "delete_xla_buffers",
]



import pandas as pd
import plotly.express as px
from PIL import (
    Image,
    ImageDraw,
)

from ..cui.common import load
from ..spike import get_dynamics
from .common import to_image


def spike_image(
    cfg, stage, tsel=slice(None), ksel=slice(None), width=3, lines=(), thr_udense=1.0
):
    pad = get_dynamics(cfg.dynamics).size - 1
    u, _, _ = load(cfg, "temporal", stage)
    u = u[:, pad:]
    stats = load(cfg, "evaluate", stage)
    stats = stats.query("kind == 'cell'")
    #u = u[stats.udense <= thr_udense]
    return _spike_image(u, tsel, ksel, width, lines)


def _spike_image(u, tsel=slice(None), ksel=slice(None), width=3, lines=()):
    u /= u.max(axis=1, keepdims=True)
    u = u[ksel, tsel]
    nk, nt = u.shape
    img = to_image(u, "Reds")
    img = Image.fromarray(img)
    draw = ImageDraw.Draw(img)
    draw.line(
        (int(0.8 * nt), (nk - 10), int(0.9 * nt), (nk - 10)),
        fill=(0, 0, 0),
        width=width,
    )
    for k0, k1, color, width in lines:
        if k0 < 0:
            k0 = nk + k0
        if k1 < 0:
            k1 = nk + k1
        draw.line((0, k0, 0, k1), fill=color, width=width)
    return img, nt, nk


def spike_stats_fig(cfg, stages, **kwargs):
    kwargs.setdefault("template", "none")
    kwargs.setdefault("margin", dict(l=40, r=10, t=20, b=35))

    dfs = []
    for stage in stages:
        stats, _ = load(cfg, "evaluate", stage)
        stats["epoch"] = stage
        dfs.append(stats)
    stats = pd.concat(dfs, axis=0)
    fig = px.scatter(
        stats,
        x="signal",
        y="udense",
        opacity=0.3,
        facet_col="epoch",
        labels=dict(signal="signal intensity", udense="spike density"),
    )
    fig.update_layout(
        **kwargs,
    )
    return fig



from functools import partial

import jax
import jax.lax as lax
import jax.numpy as jnp
import numpy as np


@partial(jax.jit, static_argnames=["r"])
def gaussian(imgs, r):
    return _gaussian(imgs, r, 4 * np.ceil(r))


def _gaussian(imgs, r, nd):
    sqrt_2pi = jnp.sqrt(2 * jnp.pi)
    d = jnp.square(jnp.arange(-nd, nd + 1, 1))
    r2 = jnp.square(r)
    o0 = jnp.exp(-d / r2 / 2) / r / sqrt_2pi
    g = imgs[..., None, :, :]
    g = lax.conv(g, o0[None, None, :, None], (1, 1), "same")
    g = lax.conv(g, o0[None, None, None, :], (1, 1), "same")
    return g[..., 0, :, :]



from logging import getLogger

import numpy as np
import pandas as pd

from ..utils.math import (
    calc_sn,
    robust_zscore,
    calc_mah,
)
from .dynamics import (
    get_dynamics,
    get_rdynamics,
)

logger = getLogger(__name__)


def evaluate(stats, spikes, bg):
    sm = spikes.max(axis=1)
    sd = spikes.mean(axis=1) / sm
    nonzero = np.count_nonzero(spikes > 0, axis=1)
    sn = calc_sn(spikes)
    rsn = 1 / sn
    zrsn = robust_zscore(rsn)

    cond_cell = stats.kind == "cell"
    x = stats.loc[cond_cell, "firmness"].to_numpy()
    y = rsn
    hypot = np.hypot(robust_zscore(x), robust_zscore(y))
    mah = calc_mah(x, y)

    stats["spkid"] = -1
    stats.loc[cond_cell, "spkid"] = np.arange(spikes.shape[0])
    stats["signal"] = None
    stats.loc[cond_cell, "signal"] = sm
    stats["udense"] = None
    stats.loc[cond_cell, "udense"] = sd
    stats["unz"] = -1
    stats.loc[cond_cell, "unz"] = nonzero
    stats["snratio"] = None
    stats.loc[cond_cell, "snratio"] = sn
    stats["rsn"] = None
    stats.loc[cond_cell, "rsn"] = rsn
    stats["zrsn"] = None
    stats.loc[cond_cell, "zrsn"] = zrsn
    stats["hypot"] = None
    stats.loc[cond_cell, "hypot"] = hypot
    stats["mah"] = None
    stats.loc[cond_cell, "mah"] = mah

    bi = stats.kind == "background"
    bmax = np.abs(bg).max(axis=1)
    bsn = np.array([bi.max() / (1.4826 * np.median(np.abs(bi[bi != 0]))) for bi in bg])

    stats["bgid"] = -1
    stats.loc[bi, "bgid"] = list(range(np.count_nonzero(bi)))
    stats["bmax"] = None
    stats.loc[bi, "bmax"] = bmax
    stats["bsparse"] = None
    stats.loc[bi, "bsparse"] = bsn

    labels = [
        "kind",
        "old_kind",
        "segid",
        "spkid",
        "bgid",
        "y",
        "x",
        "radius",
        "aratio",
        "firmness",
        "pos_move",
        "min_dist_id",
        "min_dist",
        "max_dup_id",
        "max_dup",
        "asum",
        "area",
        "signal",
        "udense",
        "unz",
        "rsn",
        "zrsn",
        "hypot",
        "mah",
        "bmax",
        "bsparse",
    ]
    return stats[[k for k in labels if k in stats.columns]]


def fix_kind(stats, spikes, bg, dynamics, bg_type="bg", thr_bg=None, thr_cell=None):
    if thr_bg is None:
        thr_bg = {}
    if thr_cell is None:
        thr_cell = {}

    fdynamics = get_dynamics(dynamics)
    rdynamics = get_rdynamics(dynamics)
    cell_df = stats.query("kind == 'cell'")
    spikes = spikes[cell_df.spkid.to_numpy()]
    bg_df = stats.query("kind == 'background'")
    logger.info("thr %s %s", thr_bg, thr_cell)
    with np.printoptions(precision=3, suppress=True):
        bins = [-np.inf, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, np.inf]
        hist, bins = np.histogram(cell_df.zrsn, bins=bins)
        for s, e, c in zip(bins[:-1], bins[1:], hist):
            logger.info("[%f %f): %d", s, e, c)
        bins = [0.0, 0.2, 0.3, 0.35, 0.4, 0.45, 0.5, 0.6, np.inf]
        hist, bins = np.histogram(cell_df.firmness, bins=bins)
        for s, e, c in zip(bins[:-1], bins[1:], hist):
            logger.info("[%f %f): %d", s, e, c)

    if bg_type == "remove":
        remove_mask = np.zeros(cell_df.shape[0], bool)
        for k, thr in thr_bg.items():
            if k in ["rsn", "zrsn", "hypot", "mah"]:
                remove_mask |= cell_df[k] > thr
            else:
                remove_mask |= cell_df[k] < thr
        spikes = spikes[~remove_mask]
        cell_df = cell_df.loc[~remove_mask].copy()
        cell_df["spkid"] = np.arange(cell_df.shape[0])

        stats = cell_df
    elif bg_type == "bg":
        to_bg_mask = np.zeros(cell_df.shape[0], bool)
        for k, thr in thr_bg.items():
            if k in ["rsn", "zrsn", "hypot", "mah"]:
                to_bg_mask |= cell_df[k] > thr
            else:
                to_bg_mask |= cell_df[k] < thr

        if len(thr_cell) == 0:
            to_cell_mask = np.zeros(bg_df.shape[0], bool)
        else:
            to_cell_mask = np.ones(bg_df.shape[0], bool)
            for k, thr in thr_cell.items():
                to_cell_mask &= bg_df[k] > thr

        spikes, to_bg = spikes[~to_bg_mask], spikes[to_bg_mask]
        cell_df, to_bg_df = (
            cell_df.loc[~to_bg_mask].copy(), cell_df.loc[to_bg_mask].copy(),
        )
        bg, to_cell = bg[~to_cell_mask], bg[to_cell_mask]
        bg_df, to_cell_df = (
            bg_df.loc[~to_cell_mask].copy(), bg_df.loc[to_cell_mask].copy(),
        )
        to_bg = np.array(fdynamics(to_bg))
        to_cell = np.array(rdynamics(to_cell))

        spikes = np.concatenate([spikes, to_cell], axis=0)
        bg = np.concatenate([bg, to_bg], axis=0)

        nc = cell_df.shape[0]
        nb = bg_df.shape[0]
        nn = to_bg_df.shape[0]
        nm = to_cell_df.shape[0]

        cell_df["spkid"] = np.arange(cell_df.shape[0])
        bg_df["bgid"] = np.arange(bg_df.shape[0])

        to_bg_df["kind"] = "background"
        to_bg_df["spkid"] = -1
        to_bg_df["bgid"] = np.arange(nb, nb + nn, dtype=np.int32)

        to_cell_df["kind"] = "cell"
        to_cell_df["spkid"] = np.arange(nc, nc + nm, dtype=np.int32)
        to_cell_df["bgid"] = -1

        stats = pd.concat([cell_df, to_cell_df, bg_df, to_bg_df], axis=0)
    else:
        raise ValueError()
    return stats, spikes, bg



from pathlib import Path

import hydra

from hotaru.cui.common import (
    all_stats,
    get_data,
)
from hotaru.plot.densesig import dense_sig_fig
from hotaru.plot.seg import (
    bg_sum_image,
    seg_max_fig,
    seg_max_image,
    segs_image,
)
from hotaru.plot.spike import spike_image
from hotaru.plot.num import cell_num_fig


@hydra.main(version_base=None, config_path="pkg://hotaru.conf", config_name="config")
def main(cfg):
    name = f"{cfg.data.label}-{cfg.penalty.ulabel}-{cfg.clean.label}"

    final = len(all_stats(cfg)) - 1
    # final = 15
    h, w = get_data(cfg).shape

    fig_dir = Path("figs")
    fig_dir.mkdir(parents=True, exist_ok=True)

    """
    footprint_stats_fig(cfg, [0], usefind=True, width=200, height=200).write_image(
        fig_dir / f"{name}0peak.pdf",
    )
    seg_max_image(cfg, 0, base=0.5).save(
        fig_dir / f"{name}0segmax.png",
        dpi=(w / 3, w / 3),
    )
    bg_sum_image(cfg, 0).save(
        fig_dir / f"{name}0bgsum.png",
        dpi=(w / 3, w / 3),
    )

    segs_image(cfg, 0, range(0, 100), mx=10, hsize=25, pad=2).save(
        fig_dir / f"{name}0segtop.png",
        dpi=(255, 255),
    )
    segs_image(cfg, 0, range(-100, 0), mx=10, hsize=25, pad=2).save(
        fig_dir / f"{name}0segbottom.png",
        dpi=(255, 255),
    )
    img, nt, nk = spike_image(
        cfg, 0, tsel=slice(t0:=400, t1:=800),
        lines=((0, 99, (0, 0, 0), 3), (-100, -1, (0, 0, 0), 3)),
    )
    img.save(
        fig_dir / f"{name}0spike.png",
        dpi=(int(0.3 * nt), int(0.3 * nk)),
    )
    """

    thr_udense = 0.18

    """
    cell_num_fig(cfg, width=400, height=200).write_image(
        fig_dir / f"{name}run_num.pdf",
    )
    dense_sig_fig(
        cfg, [0, 1, final], thr_sig=cfg.clean.args.thr_bg_signal, label="epoch="
    ).write_image(
        fig_dir / f"{name}densesig.pdf",
    )
    """

    seg_max_fig(cfg, final, base=0.2, width=300, thr_udense=thr_udense).write_image(
        fig_dir / f"{name}finalsegmax.pdf",
    )
    seg_max_image(cfg, final, base=0.2).save(
        fig_dir / f"{name}finalsegmax.png",
        dpi=(w / 3, w / 3),
    )
    bg_sum_image(cfg, final).save(
        fig_dir / f"{name}finalbgsum.png",
        dpi=(w / 3, w / 3),
    )
    segs_image(cfg, final, mx=10, hsize=25, pad=2, thr_udense=thr_udense).save(
        fig_dir / f"{name}finalseg.png",
        dpi=(255, 255),
    )

    img, nt, nk = spike_image(
        cfg,
        final,
        #tsel=slice(t0 := 400, t1 := 800),
        thr_udense=thr_udense,
    )
    img.save(
        fig_dir / f"{name}finalspike.png",
        dpi=(int(0.3 * nt), int(0.3 * nk)),
    )

    """
    run_fig(
        cfg,
        [1, final],
        thr_f=cfg.clean.args.thr_bg_firmness,
        thr_d=cfg.clean.args.thr_bg_udense,
        min_val=(0.2, 0),
        max_val=(0.58, 0.2),
        min_r=1,
        max_r=1,
    ).write_image(
        fig_dir / f"{name}run.pdf",
    )
    """
    """
    spike_stats_fig(cfg, [0, 1, final], width=600, height=200).write_image(
        fig_dir / f"{name}run_spike.pdf",
    )
    footprint_stats_fig(cfg, [0, 1, final], width=600, height=200).write_image(
        fig_dir / f"{name}run_footprint.pdf",
    )
    """


main()



from collections import namedtuple
from logging import getLogger

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf

from ..filter import (
    gaussian,
    gaussian_laplace,
    max_pool,
)
from ..utils import (
    from_tf,
    get_gpu_env,
)
from .radius import get_radius

logger = getLogger(__name__)

PeakVal = namedtuple("PeakVal", ["radius", "t", "r", "v"])


def find_peaks(data, radius, env=None, factor=1, prefetch=1):
    @jax.jit
    def update(ts, rs, gs, index, imgs):
        t, r, g = _find_peaks(imgs, mask, radius)
        cond = g < gs
        ts = jnp.where(cond, ts, index[t])
        rs = jnp.where(cond, rs, r)
        gs = jnp.where(cond, gs, g)
        return ts, rs, gs

    nt, h, w = data.imgs.shape

    radius = get_radius(radius)
    env = get_gpu_env(env)
    nd = env.num_devices
    batch = env.batch(float(factor) * h * w * len(radius), nt)
    sharding = env.sharding((nd, 1))

    logger.info(
        "find: nt=%d h=%d w=%d rmin=%f rmax=%f batch=%d",
        nt,
        h,
        w,
        radius[0],
        radius[-1],
        batch,
    )
    dataset = tf.data.Dataset.from_generator(
        lambda: zip(range(nt), data.data(mask_type=False)),
        output_signature=(
            tf.TensorSpec((), tf.int32),
            tf.TensorSpec((h, w), tf.float32),
        ),
    )
    dataset = dataset.batch(batch)
    dataset = dataset.prefetch(prefetch)

    mask = None if data.mask is None else jnp.array(data.mask, bool)
    ts = jnp.full((h, w), -1, jnp.int32)
    rs = jnp.full((h, w), -1, jnp.int32)
    gs = jnp.full((h, w), -jnp.inf)

    logger.info("%s: %s %s %d", "pbar", "start", "find", nt)
    for d in dataset:
        d = (from_tf(v) for v in d)
        index, imgs = (jax.device_put(v, sharding) for v in d)

        count = index.size
        diff = batch - count
        if diff > 0:
            index = jnp.pad(index, ((0, diff)), constant_values=-1)
            imgs = jnp.pad(imgs, ((0, diff), (0, 0), (0, 0)), constant_values=jnp.nan)

        ts, rs, gs = update(ts, rs, gs, index, imgs)
        logger.info("%s: %s %d", "pbar", "update", count)
    logger.info("%s: %s", "pbar", "close")

    for i, r in enumerate(radius):
        logger.info("radius=%f num=%d", r, (rs == i).sum())
    return PeakVal(np.array(radius, np.float32), *map(np.array, (ts, rs, gs)))


def simple_peaks(img, gauss, maxpool):
    g = gaussian(img[None, ...], gauss)[0]
    m = max_pool(g, (maxpool, maxpool), (1, 1), "same")
    y, x = jnp.where(g == m)
    return np.array(y), np.array(x)


def simple_find(imgs, mask, radius):
    t, r, v = (np.array(o) for o in _find_peaks(imgs, mask, get_radius(**radius)))
    idx = jnp.where(np.isfinite(v))
    return t[idx], idx[:, 0], idx[:, 1], r[idx], v[idx]


def _find_peaks(imgs, mask, radius):
    nt, h, w = imgs.shape
    nr = len(radius)
    gl = gaussian_laplace(imgs, radius, axis=1)
    gl_max = max_pool(gl, (3, 3, 3), (1, 1, 1), "same")
    gl_peak = gl == gl_max
    if mask is not None:
        gl_peak &= mask
    gl = jnp.where(gl_peak, gl, -jnp.inf)
    gl_reshape = gl.reshape(nt * nr, h, w)
    idx = jnp.argmax(gl_reshape, axis=0)
    gl_max = jnp.take_along_axis(gl_reshape, idx[jnp.newaxis, ...], axis=0)[0]
    t, r = jnp.divmod(idx, nr)
    return t, r, gl_max



from logging import getLogger

import jax
import jax.numpy as jnp
import tensorflow as tf

logger = getLogger(__name__)


def mapped_imgs(dataset, nt, calc, types, init, batch, pad_value=jnp.nan, in_axes=0, args=()):
    @jax.jit
    def aggrigate(val, out):
        for i, k in enumerate(types):
            match k:
                case "min":
                    vali = val[i].min(axis=0)
                    out[i] = jnp.minimum(out[i], vali)
                case "max":
                    vali = val[i].max(axis=0)
                    out[i] = jnp.maximum(out[i], vali)
                case "add":
                    vali = val[i].sum(axis=0)
                    out[i] += vali
                case ("stack", index):
                    s0, s1, *shape = val[i].shape
                    vali = val[i].reshape([s0 * s1] + shape)
                    out[i] = out[i].at[val[index].ravel()].set(vali)
                case ("argmax", index):
                    idx = jnp.argmax(val[-1], axis=0, keepdims=True)
                    val0 = jnp.max(val[k[1]], axis=0)
                    vali = jnp.take_along_axis(val[i], idx, axis=0)[0]
                    out[i] = jnp.where(out[index] > val0, out[i], vali)
                case "nouse":
                    pass
        return out

    value_map = {
        tf.float32: pad_value,
        tf.uint16: tf.constant(0, tf.uint16),
        tf.int16: tf.constant(0, tf.int16),
        tf.int32: -1,
        tf.int64: -1,
    }

    calc = jax.pmap(calc, in_axes=in_axes)

    spec = dataset.element_spec
    batch_size = batch[0] * batch[1]
    shapes = tuple((batch[1], *s.shape.as_list()) for s in spec)
    values = tuple(value_map[s.dtype] for s in spec)

    dataset = dataset.batch(batch[1])
    dataset = dataset.padded_batch(batch[0], shapes, values)
    dataset = dataset.prefetch(10)
    #logger.debug("batch %s", batch)
    #logger.debug("shapes %s", shapes)
    #logger.debug("values %s", values)

    end = 0
    out = init
    for clip in dataset.as_numpy_iterator():
        start, end = end, end + batch_size
        clip = [jnp.array(c) for c in clip]
        val = calc(*clip, *args)
        #logger.debug("val/out %s/%s", [v.shape for v in val], [o.shape for o in out])
        out = aggrigate(val, out)
        n = batch_size if end < nt else nt - start
        logger.info("%s: %s %d", "pbar", "update", n)
    return out



import jax.lax as lax
import jax.numpy as jnp


def get_segment_mask(val, y0, x0):
    def cond(args):
        seg, old_seg = args
        return jnp.any(seg != old_seg)

    def body(args):
        seg, old_seg = args
        old_seg = seg.copy()
        for dy, dx in delta:
            nseg = jnp.roll(seg, (dy, dx), axis=(0, 1))
            if dy == 1:
                nseg = nseg.at[0, :].set(False)
            if dy == -1:
                nseg = nseg.at[-1, :].set(False)
            if dx == 1:
                nseg = nseg.at[:, 0].set(False)
            if dx == -1:
                nseg = nseg.at[:, -1].set(False)
            nval = jnp.roll(val, (dy, dx), axis=(0, 1))
            nseg &= (0 < val) & (val <= nval)
            seg |= nseg
        return seg, old_seg

    delta = [(dy, dx) for dy in [-1, 0, 1] for dx in [-1, 0, 1] if (dy, dx) != (0, 0)]

    old_seg = jnp.zeros_like(val, bool)
    seg = jnp.zeros_like(val, bool)
    seg = seg.at[y0, x0].set(True)
    init = seg, old_seg
    return lax.while_loop(cond, body, init)[0]



import dash.dcc as dcc
import dash.html as html
import dash_bootstrap_components as dbc
from dash import (
    Input,
    Output,
    Patch,
    State,
    callback,
)
from omegaconf import OmegaConf

from ..footprint.find import simple_peaks
from ..main import stats
from .graph import (
    img_fig,
    scatter_fig,
    update_img_fig,
    update_scatter_fig,
)
from .progress import Progress


def create_stats_div(cfg_store):
    gaussian = dcc.Slider(
        min=0,
        max=10,
        step=0.1,
        value=1,
        marks={i: str(i) for i in range(11)},
    )
    maxpool = dcc.Slider(
        min=3,
        max=11,
        step=2,
        value=3,
    )

    std_graph = dcc.Graph(figure=img_fig("std"))
    cor_graph = dcc.Graph(figure=img_fig("cor"))
    std_cor_graph = dcc.Graph(figure=scatter_fig("", "std", "cor"))

    status_div = html.Div(
        style=dict(
            # width="1200px",
            display="grid",
            gridTemplateColumns="auto 100px 1fr",
        ),
        children=[
            get_stats := dbc.Button("STATS"),
            session := html.Div(),
            progress := dbc.Progress(value="0"),
        ],
    )

    control_div = html.Div(
        style=dict(
            # width="1200px",
            display="grid",
            gridTemplateColumns="auto 1fr",
        ),
        children=[
            dbc.Label("Gaussian"),
            gaussian,
            dbc.Label("MaxPool"),
            maxpool,
        ],
    )

    graph_div = html.Div(
        style=dict(
            # width="1200px",
            display="grid",
            columnGap="10px",
            gridTemplateColumns="auto auto auto",
        ),
        children=[
            std_graph,
            cor_graph,
            std_cor_graph,
        ],
    )

    stats_div = html.Div(
        style=dict(
            width="1200px",
        ),
        children=[
            status_div,
            control_div,
            graph_div,
        ],
    )

    @callback(
        Output(std_graph, "figure"),
        Output(cor_graph, "figure"),
        Output(std_cor_graph, "figure"),
        Input(get_stats, "n_clicks"),
        Input(gaussian, "value"),
        Input(maxpool, "value"),
        State(cfg_store, "data"),
        background=True,
        interval=100,
        running=[
            (Output(get_stats, "disabled"), True, False),
            (Output(gaussian, "disabled"), True, False),
            (Output(maxpool, "disabled"), True, False),
        ],
        progress=[
            Output(session, "children"),
            Output(progress, "value"),
            Output(progress, "max"),
            Output(progress, "label"),
        ],
        prevent_initial_call=True,
    )
    def get_stats(set_progress, n_cilcks, gaussian, maxpool, cfg):
        pbar = Progress(set_progress)
        _cfg = OmegaConf.create(cfg)
        _stats = stats(_cfg, pbar)
        y, x = simple_peaks(_stats.icor, gaussian, maxpool, pbar.session("peaks"))
        std_fig = update_img_fig(Patch(), _stats.istd, x, y)
        cor_fig = update_img_fig(Patch(), _stats.icor, x, y)
        std_cor_fig = update_scatter_fig(Patch(), _stats.istd, _stats.icor, x, y)
        return std_fig, cor_fig, std_cor_fig

    return stats_div



from logging import getLogger

import jax.numpy as jnp
import numpy as np

from .common import (
    loss_fn,
    prepare_matrix,
)
from ..spike import get_dynamics
from .optimizer import ProxOptimizer
from .penalty import get_penalty
from .regularizer import (
    L1,
    NonNegativeL1,
)

logger = getLogger(__name__)


class Model:
    def __init__(self, data, trans, stats, dynamics, penalty, **kwargs):
        self._dynamics = get_dynamics(dynamics)
        self._penalty = get_penalty(penalty)

        self._data = data
        self._trans = trans
        self._stats = stats

    @property
    def regularizers(self):
        raise NotImplementedError()

    def _try_clip(self, clip, segs):
        cdf = self._stats.query("kind=='cell'")
        bdf = self._stats.query("kind=='background'")
        cid = cdf.segid.to_numpy().astype(np.int32)
        bid = bdf.segid.to_numpy().astype(np.int32)
        fp = segs[cid]
        bg = segs[bid]

        clipped_fp = clip.clip(fp)
        clipped1 = np.any(clipped_fp > 0, axis=(1, 2))
        clipped_cdf = cdf[clipped1]
        clipped_fp = clipped_fp[clipped1]

        clipped_bg = clip.clip(bg)
        clipped2 = np.any(clipped_bg > 0, axis=(1, 2))
        clipped_bdf = bdf[clipped2]
        clipped_bg = clipped_bg[clipped2]

        active1 = clip.in_clipping_area(clipped_cdf.y, clipped_cdf.x)
        active2 = clip.in_clipping_area(clipped_bdf.y, clipped_bdf.x)

        self._clip = clip
        self._active = active1, active2

        index1 = np.nonzero(clipped1)[0][active1]
        index2 = np.nonzero(clipped2)[0][active2]
        self._active_index = index1, index2

        return clipped1, clipped2, clipped_fp, clipped_bg

    def _prepare(self, data, yval, py, **kwargs):
        ycov, yout, ydot = prepare_matrix(data, yval, self._trans, **kwargs)

        penalty = self._penalty
        if self._trans:
            nx, ny = data.nt, data.ns
            bx, by = penalty.bt, penalty.bs
            logger.info("lu: %f", self._penalty.lu[1][0])
        else:
            nx, ny = data.ns, data.nt
            bx, by = penalty.bs, penalty.bt
            logger.info("la: %f", self._penalty.la[1][0])
        nx, ny, bx, by = (jnp.array(v, jnp.float32) for v in (nx, ny, bx, by))

        self._args = ycov, yout, ydot, nx, ny, bx, by, py
        self._loss_scale = nx * ny + nx + ny
        self._penalty_scale = ny

        est = np.linalg.eigh(ycov)[0].max()
        if not np.isfinite(est):
            est = 1.0
        self._lr_scale = est / nx

        if not hasattr(self, "_optimizer"):
            self._optimizer = ProxOptimizer(self)

    @property
    def args(self):
        return self._args

    @property
    def loss_scale(self):
        return self._loss_scale

    def loss(self, x1, x2, ycov, yout, ydot, nx, ny, bx, by, py):
        x = jnp.concatenate([x1, x2], axis=0)
        loss, sigma = loss_fn(x, ycov, yout, ydot, nx, ny, bx, by)
        return loss + py, sigma

    @property
    def prox(self):
        return (r.prox for r in self.regularizers)

    def fit(self, max_epoch, steps_par_epoch, lr, *args, **kwargs):
        logger.info("fit: lr=%g scale=%f", lr, self._lr_scale)
        lr /= self._lr_scale
        x = self._x
        logger.info("%s: %s %s %d", "pbar", "start", "optimize", -1)
        x, history = self._optimizer.fit(
            x, max_epoch, steps_par_epoch, lr, *args, **kwargs
        )
        logger.info("%s: %s", "pbar", "close")
        self._x = x
        return history

    def get_x(self):
        index1, index2 = self._active_index
        x1, x2 = self._x
        active1, active2 = self._active
        active_x1 = np.array(x1[active1])
        active_x2 = np.array(x2[active2])
        return index1, index2, active_x1, active_x2


class SpatialModel(Model):
    def __init__(self, data, stats, oldx, y1, y2, *args, **kwargs):
        super().__init__(data, False, stats, *args, **kwargs)
        self._oldx = oldx
        self._y1 = y1
        self._y2 = y2

    @property
    def n1(self):
        return self._y1.shape[0]

    @property
    def n2(self):
        return self._y2.shape[0]

    @property
    def regularizers(self):
        return self._penalty.la[0], NonNegativeL1()

    @property
    def prox_args(self):
        return (self._penalty_scale * self._penalty.la[1][0],), (self._lb,)

    def try_clip(self, clip):
        return self._try_clip(clip, self._oldx)

    def prepare(self, clip, **kwargs):
        data = self._data.clip(clip.clip)
        clipped1, clipped2, clipped_img1, clipped_img2 = self.try_clip(clip)

        clipped_y1 = jnp.array(self._y1[clipped1])
        clipped_y2 = jnp.array(self._y2[clipped2])

        lu, fac = self._penalty.lu
        py = data.nt * lu(clipped_y1, *fac)
        self._lb = jnp.abs(self._penalty.lb * clipped_y2).sum(axis=1, keepdims=True)

        clipped_z1 = self._dynamics(clipped_y1)
        clipped_z2 = clipped_y2
        clipped_z1 /= clipped_z1.max(axis=1, keepdims=True)
        clipped_z2 /= clipped_z2.max(axis=1, keepdims=True)
        zval = jnp.concatenate([clipped_z1, clipped_z2], axis=0)

        self._prepare(data, zval, py, **kwargs)

        clipped_x1 = jnp.array(data.apply_mask(clipped_img1, mask_type=True))
        clipped_x2 = jnp.array(data.apply_mask(clipped_img2, mask_type=True))
        self._x = clipped_x1, clipped_x2

    def get_x(self):
        index1, index2, x1, x2 = super().get_x()

        index = np.concatenate([index1, index2 + self.n1], axis=0)
        x = np.concatenate([x1, x2], axis=0)

        mask = self._data.mask
        shape = self._data.shape
        x = self._clip.unclip(x, mask, shape)
        return index, x


class TemporalModel(Model):
    def __init__(self, data, stats, y, *args, **kwargs):
        super().__init__(data, True, stats, *args, **kwargs)
        self._y = y

    @property
    def n1(self):
        return np.count_nonzero(self._stats.kind == "cell")

    @property
    def n2(self):
        return np.count_nonzero(self._stats.kind == "background")

    @property
    def regularizers(self):
        return self._penalty.lu[0], L1()

    @property
    def prox_args(self):
        return (self._penalty_scale * self._penalty.lu[1][0],), (self._lb,)

    def try_clip(self, clip):
        return self._try_clip(clip, self._y)

    def prepare(self, clip, **kwargs):
        _, _, clipped_y1, clipped_y2 = self.try_clip(clip)
        data = self._data.clip(clip.clip)

        clipped_y1 = data.apply_mask(clipped_y1, mask_type=True)
        clipped_y2 = data.apply_mask(clipped_y2, mask_type=True)

        clipped_y1 = jnp.array(clipped_y1)
        clipped_y2 = jnp.array(clipped_y2)

        la, fac = self._penalty.la
        py = la(clipped_y1, *fac)

        lb = self._penalty.lb
        self._lb = jnp.abs(lb * clipped_y2).sum(axis=1, keepdims=True)

        yval = jnp.concatenate([clipped_y1, clipped_y2], axis=0)
        self._prepare(data, yval, py, **kwargs)

        active1, active2 = self._active
        n1, n2 = active1.size, active2.size
        nt = self._data.nt
        nu = nt + self._dynamics.size - 1
        x1 = jnp.zeros((n1, nu))
        x2 = jnp.zeros((n2, nt))
        self._x = x1, x2

    def loss(self, x1, x2, ycov, yout, ydot, nx, ny, bx, by, py):
        x1 = self._dynamics(x1)
        return super().loss(x1, x2, ycov, yout, ydot, nx, ny, bx, by, py)



from collections import namedtuple

import numpy as np


def get_clip(shape, clip):
    h, w = shape
    match clip:
        case [Clip(), *_]:
            return clip
        case None:
            return [Clip(0, h, 0, w, 0)]
        case {"type": "Div", "ynum": ynum, "xnum": xnum, "margin": margin}:
            dy = (h - 2 * margin + ynum - 1) // ynum
            dx = (w - 2 * margin + xnum - 1) // xnum
            clips = []
            for py in range(ynum):
                if py == 0:
                    y0 = 0
                else:
                    y0 = margin + dy * py
                if py == ynum - 1:
                    y1 = h
                else:
                    y1 = margin + dy * (py + 1)
                for px in range(xnum):
                    if px == 0:
                        x0 = 0
                    else:
                        x0 = margin + dx * px
                    if px == xnum - 1:
                        x1 = w
                    else:
                        x1 = margin + dx * (px + 1)
                    clips.append(Clip(y0, y1, x0, x1, margin))
            return clips
        case _:
            raise ValueError()


class Clip(namedtuple("Clip", "y0 y1 x0 x1 margin")):
    def range(self, h, w):
        y0 = max(self.y0 - self.margin, 0)
        y1 = min(self.y1 + self.margin, h)
        x0 = max(self.x0 - self.margin, 0)
        x1 = min(self.x1 + self.margin, w)
        return y0, y1, x0, x1

    def clip(self, img):
        if img is None:
            return img
        else:
            h, w = img.shape[-2:]
            y0, y1, x0, x1 = self.range(h, w)
            return img[..., y0:y1, x0:x1]

    def unclip(self, val, mask, shape):
        nk = val.shape[0]
        h, w = shape
        y0, y1, x0, x1 = self.range(h, w)
        if mask is None:
            tmp = val.reshape(nk, y1 - y0, x1 - x0)
        else:
            tmp = np.zeros((nk, y1 - y0, x1 - x0), val.dtype)
            tmp[:, mask] = val
        out = np.zeros((nk, h, w), val.dtype)
        out[:, y0:y1, x0:x1] = tmp
        return out

    def in_clipping_area(self, y, x):
        return (self.y0 <= y) & (y < self.y1) & (self.x0 <= x) & (x < self.x1)



import jax.lax as lax
import jax.numpy as jnp
import numpy as np


def get_dynamics(dynamics):
    match dynamics:
        case SpikeToCalcium():
            return dynamics
        case {"label": _, "type": "double_exp", **args}:
            return SpikeToCalcium.double_exp(**args)
        case _:
            raise ValueError("invalid dynamics type")


def get_rdynamics(dynamics):
    match dynamics:
        case CalciumToSpike():
            return dynamics
        case {"label": _, "type": "double_exp", **args}:
            return CalciumToSpike.double_exp(**args)
        case _:
            raise ValueError("invalid dynamics type")


class SpikeToCalcium:
    @classmethod
    def double_exp(cls, tau1, tau2, duration, hz=1.0):
        if tau1 > tau2:
            tau1, tau2 = tau2, tau1
        size = int(np.ceil(duration * hz))
        tau1 *= hz
        tau2 *= hz
        r = tau1 / tau2
        d = tau1 - tau2
        scale = np.power(r, -tau2 / d) - np.power(r, -tau1 / d)
        t = np.arange(1, size + 1)
        e1 = np.exp(-t / tau1)
        e2 = np.exp(-t / tau2)
        return cls((e1 - e2) / scale)

    def __init__(self, kernel):
        self.kernel = jnp.array(kernel, jnp.float32)[jnp.newaxis, jnp.newaxis, ::-1]

    def __call__(self, spike):
        spike = spike[:, jnp.newaxis, ...]
        return lax.conv(spike, self.kernel, (1,), "valid")[:, 0, ...]

    @property
    def size(self):
        return self.kernel.size


class CalciumToSpike:
    @classmethod
    def double_exp(cls, tau1, tau2, duration, hz=1.0):
        if tau1 > tau2:
            tau1, tau2 = tau2, tau1
        tau1 *= hz
        tau2 *= hz
        r = tau1 / tau2
        d = tau1 - tau2
        scale = np.power(r, -tau2 / d) - np.power(r, -tau1 / d)
        e1 = np.exp(-1 / tau1)
        e2 = np.exp(-1 / tau2)
        kernel = np.array([1.0, -(e1 + e2), e1 * e2]) / (e1 - e2) * scale
        size = int(np.ceil(duration * hz))
        pad = size + 1
        return cls(kernel, pad)

    def __init__(self, kernel, pad):
        self.kernel = jnp.array(kernel, jnp.float32)[jnp.newaxis, jnp.newaxis, ::-1]
        self.pad = pad

    def __call__(self, calcium):
        calcium = calcium[:, jnp.newaxis, ...]
        spike = lax.conv(calcium, self.kernel, (1,), "valid")[:, 0, ...]
        spike = jnp.pad(spike, [[0, 0], [self.pad, 0]])
        spike += spike.min(axis=1, keepdims=True)
        return spike

    @property
    def size(self):
        return self.kernel.size



import numpy as np
import plotly.graph_objects as go
from PIL import Image

from ..cui.common import load
from ..footprint import get_radius
from .common import to_image


def seg_max_image(cfg, stage, base=0.5, showbg=False, thr_udense=1.0):
    if stage == 0:
        segs = load(cfg, "make", stage)
    else:
        segs, _ = load(cfg, "clean", stage)
    stats, _ = load(cfg, "evaluate", stage)
    return _seg_max_image(segs, stats, base, showbg, thr_udense)


def _seg_max_image(segs, stats, base=0.5, showbg=False, thr_udense=1.0):
    #stats.loc[stats.udense > thr_udense, "kind"] = "background"
    cell = stats.query("kind == 'cell'").segid.to_numpy()
    bg = stats.query("kind == 'background'").segid.to_numpy()
    fp = segs[cell]
    fp = np.maximum(0, (fp - base) / (1 - base)).max(axis=0)
    fpimg = to_image(fp, "Greens")
    fpimg = Image.fromarray(fpimg)
    if showbg:
        bg = segs[bg].max(axis=0)
        bgimg = to_image(bg, "Reds")
        bgimg[:, :, 3] = 128
        fpimg = Image.fromarray(bgimg)
        fpimg.paset(bgimg, (0, 0), bgimg)
    return fpimg


def seg_max_fig(cfg, stage, base=0.5, showbg=False, width=600, thr_udense=1.0):
    if stage == 0:
        segs = load(cfg, "make", stage)
        stats = load(cfg, "init", stage)
    else:
        _, segs = load(cfg, "clean", stage)
        stats = load(cfg, "evaluate", stage)
    return _seg_max_fig(segs, stats, base, showbg, width, thr_udense)


def _seg_max_fig(segs, stats, base=0.5, showbg=False, width=600, thr_udense=1.0):
    img = _seg_max_image(segs, stats, base, showbg, thr_udense)
    #stats.loc[stats.udense > thr_udense, "kind"] = "background"
    cell = stats.query("kind == 'cell'")
    fig = go.Figure()
    fig.add_trace(
        go.Image(z=img),
    )
    fig.add_trace(
        go.Scatter(
            x=cell.x,
            y=cell.y,
            text=cell.index,
            mode="text",
            textfont=dict(size=5, color="red"),
        ),
    )
    w, h = img.size
    fig.update_xaxes(
        showticklabels=False,
        domain=[0, 1],
        autorange=False,
        range=(0, w),
    )
    fig.update_yaxes(
        showticklabels=False,
        domain=[0, 1],
        autorange=False,
        range=(h, 0),
    )
    fig.update_layout(
        width=width,
        height=int(width * h / w),
        margin=dict(l=0, r=0, b=0, t=0),
    )
    return fig


def bg_sum_image(cfg, stage):
    if stage == 0:
        segs = load(cfg, "make", stage)
    else:
        segs, _ = load(cfg, "clean", stage)
    stats, _ = load(cfg, "evaluate", stage)
    st = stats.query("kind == 'background'")
    bg = segs[st.index]
    bg *= st.bmax.to_numpy()[:, np.newaxis, np.newaxis]
    bgsum = bg.sum(axis=0)
    bgsum /= bgsum.max()
    img = to_image(bgsum, "Reds")
    return Image.fromarray(img)


def segs_image(cfg, stage, select=None, mx=None, hsize=20, pad=5, thr_udense=1.0):
    if stage == 0:
        segs = load(cfg, "make", stage)
        stats = load(cfg, "init", stage)
    else:
        segs, stats = load(cfg, "clean", stage)
    return _segs_image(segs, stats, select, mx, hsize, pad, thr_udense)


def _segs_image(segs, stats, select=None, mx=None, hsize=20, pad=5, thr_udense=1.0):
    def s(i):
        return pad + i * (size + pad)

    def e(i):
        return (i + 1) * (size + pad)

    size = 2 * hsize + 1

    stats.loc[stats.udense > thr_udense, "kind"] = "background"
    stats = stats.query("kind == 'cell'")

    nk = stats.shape[0]
    fp = segs[stats.index]
    ys = stats.y.to_numpy()
    xs = stats.x.to_numpy()

    if select is None:
        select = range(nk)
    else:
        fp = fp[select]
        ys = ys[select]
        xs = xs[select]
        nk = fp.shape[0]

    if mx is None:
        mx = int(np.floor(np.sqrt(nk)))
    my = (nk + mx - 1) // mx

    fp = np.pad(fp, ((0, 0), (hsize, hsize), (hsize, hsize)))
    clip = np.zeros(
        (my * size + pad * (my + 1), mx * size + pad * (mx + 1), 4), np.uint8
    )

    for x in range(mx + 1):
        st = x * (size + pad)
        en = st + pad
        clip[:, st:en] = [0, 0, 0, 255]
    for y in range(my + 1):
        st = y * (size + pad)
        en = st + pad
        clip[st:en] = [0, 0, 0, 255]

    for i, (img, y, x) in enumerate(zip(fp, ys, xs)):
        u, w = divmod(i, mx)
        clip[s(u) : e(u), s(w) : e(w)] = to_image(
            img[y : y + size, x : x + size],
            "Greens",
        )
    return Image.fromarray(clip)


def jitter(r):
    radius = np.sort(np.unique(r))
    rscale = np.log((radius[1:] / radius[:-1]).min())
    jitter = np.exp(rscale * (np.random.uniform(size=r.size) - 0.5))
    return r * jitter


def footprint_stats_fig(cfg, stages, usefind=False, **kwargs):
    kwargs.setdefault("template", "none")
    kwargs.setdefault("margin", dict(l=40, r=10, t=20, b=35))
    kwargs.setdefault("showlegend", False)

    radius = get_radius(cfg.radius.filter)
    rmin, rmax = radius[0], radius[-1]

    fig = go.Figure().set_subplots(2, len(stages), row_heights=(1, 2))
    vmax = 0
    for i, stage in enumerate(stages):
        if i == 0:
            intensity = "intensity"
            if usefind:
                peakval = load(cfg, "find", 0)
                ri = peakval.r
                cond = ri > 0
                rs = peakval.radius[ri[cond]]
                vs = peakval.v[cond]
                vmax = vs.max()
                fig.add_trace(
                    go.Scattergl(
                        x=jitter(rs),
                        y=vs,
                        mode="markers",
                        marker=dict(opacity=0.2, size=1, color="blue"),
                        name="all pixels",
                    ),
                    col=i + 1,
                    row=2,
                )
        else:
            intensity = "firmness"
        stats, _ = load(cfg, "evaluate", stage)
        cell = stats.query("kind == 'cell'")
        v = cell[intensity]
        print(intensity, v)
        vmax = max(v.max(), vmax)
        fig.add_trace(
            go.Scattergl(
                x=cell.radius if usefind else jitter(cell.radius),
                y=v,
                mode="markers",
                marker=dict(opacity=0.3, size=5, color="green"),
                name="all pixels",
            ),
            col=i + 1,
            row=2,
        )
        r, c = np.unique(cell.radius, return_counts=True)
        print(r)
        print(c)
        fig.add_trace(
            go.Bar(
                x=np.log(r),
                y=c,
                marker_color="green",
            ),
            col=i + 1,
            row=1,
        )
        fig.update_xaxes(
            showticklabels=False,
            # tickmode="array",
            # tickvals=[np.log(2), np.log(4), np.log(8)],
            # ticktext=["2", "4", "8"],
            range=[np.log(rmin), np.log(rmax)],
            col=i + 1,
            row=1,
        )
        print(rmin, rmax)
        fig.update_xaxes(
            title_text="radius",
            type="log",
            tickmode="array",
            tickvals=[3, 6, 12],
            ticktext=["3", "6", "12"],
            autorange=False,
            range=[np.log10(rmin), np.log10(rmax)],
            col=i + 1,
            row=2,
        )
        if i == 0:
            fig.update_yaxes(
                title_text="intensity",
                col=1,
                row=2,
                range=(0, 1.05 * vmax),
            )
            vmax = 0
    fig.update_yaxes(
        title_text="fimness",
        col=2,
        row=2,
    )
    print(vmax)
    for i in range(1, len(stats)):
        fig.update_yaxes(
            range=(0, 1.05 * vmax),
            col=i + 1,
            row=2,
        )
    fig.update_yaxes(
        title_text="count",
        col=1,
        row=1,
    )
    fig.update_layout(**kwargs)
    return fig



from logging import getLogger

import numpy as np

from ..footprint import (
    find_peaks,
    make_footprints,
    reduce_peaks,
)
from ..io import (
    save,
    try_load,
)
from ..utils import get_xla_stats
from .common import (
    get_data,
    get_files,
    get_force,
    reduce_log,
)

logger = getLogger(__name__)


def init(cfg):
    (statsfile,) = get_files(cfg, "init", 0)
    if get_force(cfg, "init", 0) or not statsfile.exists():
        (makefile,) = get_files(cfg, "make", 0)
        (reducefile,) = get_files(cfg, "reduce", 0)
        if get_force(cfg, "make", 0) or not makefile.exists():
            logger.debug("%s", get_xla_stats())
            data = get_data(cfg)
            if get_force(cfg, "reduce", 0) or not reducefile.exists():
                (findfile,) = get_files(cfg, "find", 0)
                if get_force(cfg, "find", 0) or not findfile.exists():
                    logger.info("exec find")
                    findval = find_peaks(data, cfg.radius.filter, **cfg.cmd.find)
                    logger.debug("%s", get_xla_stats())
                    save(findfile, findval)
                    logger.info("saved find")
                else:
                    logger.info("load find")
                    findval = try_load(findfile)

                logger.info("exec reduce")
                peaks = reduce_peaks(
                    findval, bg_type=cfg.bg_type, **cfg.init.args, **cfg.cmd.reduce,
                )
                logger.debug("%s", get_xla_stats())
                save(reducefile, peaks)
                if cfg.cmd.remove_find:
                    findfile.unlink(missing_ok=True)
                logger.info("saved reduce")
            else:
                logger.info("load reduce")
                peaks = try_load(reducefile)
            reduce_log(cfg, 0)

            logger.info("exec make")
            footprints = make_footprints(data, peaks, **cfg.cmd.make)
            logger.debug("%s", get_xla_stats())
            save(makefile, footprints)
            logger.info("saved make")
        else:
            logger.info("load make")
            footprints = try_load(makefile)
            peaks = try_load(reducefile)
        logger.info("exec init")
        peaks["asum"] = footprints.sum(axis=(1, 2))
        peaks["area"] = np.count_nonzero(footprints > 0, axis=(1, 2))
        peaks.loc[peaks['asum'] == 0, 'kind'] = 'removed'
        save(statsfile, peaks)
        if cfg.cmd.remove_reduce:
            reducefile.unlink(missing_ok=True)
        logger.info("saved init")



from pathlib import Path
from logging import getLogger

import numpy as np
from tifffile import (
    TiffFile,
    imread,
    memmap,
)

logger = getLogger(__name__)


def load_imgs(**cfg):
    logger.debug("load_imgs: %s", cfg)
    match cfg["type"]:
        case "npy":
            imgs = np.load(cfg["file"], mmap_mode="r")
        case "raw":
            dtype = np.dtype(cfg["dtype"]).newbyteorder(cfg["endian"])
            data = np.memmap(cfg["file"], dtype, "r")
            imgs = data.reshape(-1, cfg["height"], cfg["width"])
        case "tif":
            imgsfile = Path(cfg["file"])
            imgsfile_fix = imgsfile.with_stem(f"{imgsfile.stem}_fix")
            if imgsfile_fix.exists():
                imgsfile = imgsfile_fix
            with TiffFile(imgsfile) as tif:
                data = tif.series[0]
                if data.dataoffset is None:
                    logger.debug("create fixed tiff: %s", imgsfile_fix)
                    imgs = memmap(imgsfile_fix, shape=data.shape, dtype=data.dtype)
                    for i, pi in enumerate(data):
                        imgs[i] = pi.asarray()
                else:
                    imgs = data.asarray(out="memmap")
            logger.debug("mount tiff as memmap: %s", imgsfile_fix)
        case _:
            raise RuntimeError(f"{imgsfile} is not imgs file")
    return imgs, cfg["hz"]


def apply_mask(imgs, **cfg):
    logger.debug("apply_mask: %s", cfg)
    match cfg["type"]:
        case "nomask":
            mask = None
        case "tif":
            mask = imread(cfg["file"])
        case "npy":
            mask = np.load(cfg["file"])
        case _:
            raise RuntimeError("bad file type: {maskfile}")

    if mask is not None:
        my = np.where(np.any(mask, axis=1))[0]
        mx = np.where(np.any(mask, axis=0))[0]
        x0, y0, w, h = mx[0], my[0], mx[-1] - mx[0] + 1, my[-1] - my[0] + 1
        imgs = imgs[:, y0 : y0 + h, x0 : x0 + w]
        mask = mask[y0 : y0 + h, x0 : x0 + w]
        logger.debug("data size: (%d, %d, %d)", *imgs.shape)

    nt, h, w = imgs.shape
    nx = -1 if mask is None else h * w
    logger.info("imgs, mask: (%d, %d, %d), %d", nt, h, w, nx)
    return imgs, mask



from logging import getLogger

from ..filter import movie_stats
from ..io import (
    apply_mask,
    load_imgs,
    save,
)
from ..utils import get_xla_stats
from .common import get_force, get_files

logger = getLogger(__name__)


def normalize(cfg):
    force = get_force(cfg, "normalize", 0)
    files = get_files(cfg, "normalize", 0)
    if force or not all(file.exists() for file in files):
        logger.info("exec normalize")
        imgs, hz = load_imgs(**cfg.data.imgs)
        imgs, mask = apply_mask(imgs, **cfg.data.mask)
        logger.info("%s", get_xla_stats())
        outputs = movie_stats(imgs, mask, **cfg.cmd.stats)
        logger.info("%s", get_xla_stats())
        save(files, outputs)
        logger.info("saved normalize")



from .laplace import (
    gaussian_laplace,
    gaussian_laplace_single,
)
from .map import mapped_imgs
from .pool import max_pool
from .stats import movie_stats

__all__ = [
    "mapped_imgs",
    "max_pool",
    "gaussian_laplace",
    "gaussian_laplace_single",
    "movie_stats",
]



import numpy as np
import jax.numpy as jnp
import plotly.graph_objects as go
import plotly.io as pio

from hotaru.filter import (
    gaussian_laplace,
    max_pool,
)
from hotaru.footprint import get_radius

pio.kaleido.scope.mathjax = None


def plot_gl(data, radius, idx, width):
    radius = get_radius(radius)
    imgs = data.select(idx)

    glrt = gaussian_laplace(imgs, radius, axis=1)
    print(glrt.shape)
    gl_max = max_pool(glrt, (3, 3, 3), (1, 1, 1), "same")
    gl_peak = glrt == gl_max

    glrt = np.array(glrt)
    gl_peak = np.array(gl_peak)
    radius = np.array(radius)

    glrt = glrt[:, 1:-1]
    gl_max = gl_max[:, 1:-1]
    gl_peak = gl_peak[:, 1:-1]
    radius = radius[1:-1]
    t, r, y, x = np.where(gl_peak & (glrt > 0))
    g = glrt[t, r, y, x]
    g /= g.max()
    r = radius[r]

    glrt = glrt[:, ::3]
    radius = radius[::3]
    nt, nr, h, w = glrt.shape
    print(nt, nr, h, w)
    print(np.stack([t, y, x, r, g], axis=1))
    print(y.size)
    #for out in zip(t, y, x, r, g):
    #    print(*out)

    fig = go.Figure().set_subplots(
        nt,
        nr + 2,
        column_widths=[w] * (nr + 2),
        row_heights=[h] * nt,
        column_titles=["image"] + [f"r={r:.1f}" for r in radius] + ["peaks"],
        row_titles=[f"t={t}" for t in idx],
        vertical_spacing=0.01,
        horizontal_spacing=0.01,
        print_grid=True,
        shared_xaxes=True,
        shared_yaxes=True,
    )
    margin = dict(l=15, r=30, t=30, b=15)
    mod_width = width - margin["l"] - margin["r"]
    for j, (img, glr) in enumerate(zip(imgs, glrt)):
        fig.add_trace(
            go.Heatmap(z=img, colorscale="Greens", showscale=False, zmin=-1, zmax=1),
            row=j + 1,
            col=1,
        )
        cond = (t == j)
        yj = y[cond]
        xj = x[cond]
        rj = r[cond]
        gj = g[cond]
        marker = dict(
            symbol="circle",
            color="blue",
            opacity=0.8 * (gj + 0.3) / 1.3,
            line_width=0,
            size=5 * rj * mod_width / (nr + 2) / w,
            sizemode="diameter",
        )
        fig.add_trace(
            go.Scatter(mode="markers", x=xj, y=yj, marker=marker),
            row=j + 1,
            col=nr + 2,
        )
        for i, gl in enumerate(glr):
            fig.add_trace(
                go.Heatmap(z=-gl, colorscale="Picnic", showscale=False, zmin=-1, zmax=1),
                row=j + 1,
                col=i + 2,
            )
    fig.update_xaxes(
        autorange=False,
        range=(0, w),
        showgrid=False,
        showticklabels=False,
    )
    fig.update_yaxes(
        autorange=False,
        range=(h, 0),
        showgrid=False,
        showticklabels=False,
    )
    fig.update_layout(
        showlegend=False,
        annotations=[
            dict(font=dict(size=11)) for _ in range(len(fig.layout.annotations))
        ],
        width=width,
        height=np.ceil(margin["t"] + margin["b"] + mod_width * nt * h / (nr + 2) / w),
        margin=margin,
    )
    return fig



import plotly.graph_objects as go


def heat_fig(title, colorscale="greens", scattercolor="red", zmin=None, zmax=None):
    heat = go.Heatmap(
        colorscale=colorscale,
        zmin=zmin,
        zmax=zmax,
        colorbar=dict(orientation="h", y=0, yanchor="top", thickness=10),
    )
    scatter = go.Scatter(mode="markers", marker_color=scattercolor)
    layout = go.Layout(
        title=title,
        xaxis=dict(visible=False, autorange=False),
        yaxis=dict(visible=False, autorange=False),
        width=600,
        height=570,
        margin=dict(t=25, b=45, l=0, r=0),
    )
    return go.Figure([heat, scatter], layout)


def spike_fig(title, hz, colorscale="reds"):
    spike = go.Heatmap(
        dx=1 / hz,
        colorscale=colorscale,
        zmin=0.0,
        zmax=1.0,
        # colorbar=dict(orientation="h", y=0, yanchor="top", thickness=10),
    )
    layout = go.Layout(
        title=title,
        xaxis=dict(title="time"),
        yaxis=dict(title="ID"),
        width=1000,
        height=570,
        margin=dict(t=25, b=45, l=0, r=0),
    )
    return go.Figure([spike], layout)


def circle_fig(title):
    scatter = go.Scatter(mode="markers", marker=dict(color="green", sizeref=10.0))
    layout = go.Layout(
        title=title,
        xaxis=dict(visible=False, autorange=False),
        yaxis=dict(visible=False, autorange=False),
        width=600,
        height=570,
        margin=dict(t=25, b=45, l=0, r=0),
    )
    return go.Figure([scatter], layout)


def update_img(fig, h, w, height=500):
    fig.layout.width = height * w / h
    fig.layout.height = height + 70
    fig.layout.xaxis.range = [0, w]
    fig.layout.yaxis.range = [h, 0]
    return fig


def scatter_fig(title, xlabel, ylabel):
    scatter = go.Scatter(mode="markers", showlegend=False)
    layout = go.Layout(
        title=title,
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        width=600,
        height=570,
    )
    return go.Figure([scatter, scatter], layout)


def bar_fig(title, xlabel, ylabel):
    bar = go.Bar()
    layout = go.Layout(
        title=title,
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        width=600,
        height=570,
    )
    return go.Figure([bar], layout)



import numpy as np
import plotly.graph_objects as go
import plotly.io as pio

from hotaru.filter import gaussian_laplace
from hotaru.footprint import get_radius

pio.kaleido.scope.mathjax = None


def plot_imgs_subplots(imgs, labels, width=500, pad=0.05):
    margin = dict(l=10, r=10, t=20, b=10)
    num = len(imgs)
    xdomainsize = (1 - (num - 1) * pad) / num
    xstride = xdomainsize + pad
    fig = go.Figure().set_subplots(1, len(imgs), subplot_titles=labels)
    for i, (img, label) in enumerate(zip(imgs, labels)):
        h, w = img.shape
        smin = img.min()
        smax = img.max()
        img = (img - smin) / (smax - smin)
        fig.add_trace(
            go.Heatmap(z=img, colorscale="Greens", showscale=False),
            col=i + 1,
            row=1,
        )
        fig.update_xaxes(
            autorange=False,
            range=(-0.5, w - 0.5),
            showticklabels=False,
            domain=[i * xstride, i * xstride + xdomainsize],
            col=i + 1,
            row=1,
        )
        fig.update_yaxes(
            scaleanchor="x",
            autorange=False,
            domain=[0, 1],
            range=(h, 0),
            showticklabels=False,
            col=i + 1,
            row=1,
        )
    mod_w = num * w * (1 + (num -1) * pad)
    mod_width = width - margin["l"] - margin["r"]
    fig.update_layout(
        annotations=[dict(font=dict(size=11)) for _ in range(num)],
        width=width,
        height=np.ceil(margin["t"] + margin["b"] + mod_width * h / mod_w),
        margin=margin,
    )
    return fig


def plot_simgs(simgs, labels, scale=1, margin=30, label=None):
    simgs = np.stack(simgs)
    h, w = simgs[0].shape
    smin = simgs.min(axis=(1, 2), keepdims=True)
    smax = simgs.max(axis=(1, 2), keepdims=True)
    simgs = (simgs - smin) / (smax - smin)
    simgs = np.pad(simgs, ((0, 0), (0, 0), (0, 1)))
    simgs = np.concatenate(simgs, axis=1)
    simgs = np.pad(simgs, ((1, 1), (0, 1)))
    kwargs = dict(
        colorscale="Greens",
        showscale=False,
    )
    margin = dict(t=margin, b=margin, l=margin, r=margin)
    data = [go.Heatmap(z=simgs, **kwargs)]
    fig = plot_clip(data, simgs.shape, h + 1, w + 1, scale, margin)
    for i, label in enumerate(labels):
        fig.add_annotation(x=10 + i * (w + 1), y=10, text=label, showarrow=False)
    return fig


def plot_gl(data, radius, idx, scale=1, margin=30, label=""):
    radius = get_radius(radius)
    imgs = data.select(idx)
    gl = gaussian_laplace(imgs, radius, 0)
    gl = np.transpose(gl, (1, 2, 0, 3))
    gl = np.pad(gl, ((0, 0), (0, 1), (0, 0), (0, 1)))
    nt, h, nr, w = gl.shape
    gl = gl.reshape(nt * h, nr * w)
    gl = np.pad(gl, ((1, 0), (1, 0)))
    kwargs = dict(
        colorscale="Picnic",
        colorbar=dict(thicknessmode="pixels", thickness=10, xpad=0, ypad=0),
        zmin=-1,
        zmax=1,
    )
    margin = dict(t=margin, b=margin, l=margin, r=margin + 40)
    data = [go.Heatmap(z=gl, **kwargs)]
    return plot_clip(data, gl.shape, h, w, scale, margin)


def plot_peak_stats(peaks, peakval=None, label=""):
    def jitter(r):
        jitter = np.exp(0.2 * dscale * np.random.randn(r.size))
        return r * jitter

    radius = np.sort(np.unique(peaks.radius))
    dscale = np.log((radius[1:] / radius[:-1]).min())
    data = []
    if peakval is None:
        intensity = "firmness"
        vmax = peaks[intensity].max()
    else:
        intensity = "intensity"
        ri = peakval.r
        cond = ri > 0
        rs = peakval.radius[ri[cond]]
        vs = peakval.v[cond]
        vmax = vs.max()
        data.append(
            go.Scattergl(
                x=jitter(rs),
                y=vs,
                mode="markers",
                marker=dict(opacity=0.03, color="blue"),
                name="all pixels",
            )
        )
    cell = peaks[peaks.kind == "cell"]
    bg = peaks[peaks.kind == "background"]
    data += [
        go.Scattergl(
            x=jitter(cell.radius),
            y=cell[intensity],
            mode="markers",
            marker=dict(size=10, symbol="star", color="lime", opacity=0.5),
            name="cell",
        ),
        go.Scattergl(
            x=jitter(bg.radius),
            y=bg[intensity],
            mode="markers",
            marker=dict(size=10, symbol="pentagon", color="gray", opacity=0.5),
            name="background",
        ),
    ]

    fig = go.Figure(
        data=data,
        layout=go.Layout(
            template="none",
            width=1000,
            height=600,
            legend=dict(
                x=0.01,
                y=0.99,
                xanchor="left",
                yanchor="top",
            ),
        ),
    )
    fig.update_xaxes(
        title="radius",
        type="log",
    )
    fig.update_yaxes(
        title=intensity,
        range=(0, 1.05 * vmax),
    )
    return fig


def plot_seg_max(
    footprints, peaks, base=0.5, plot_bg=True, scale=1, margin=30, label=""
):
    nk = np.count_nonzero(peaks.kind == "cell")
    seg = footprints[:nk]
    bg = footprints[nk:]

    margin = dict(t=margin, b=margin, l=margin, r=margin)
    nk, h, w = seg.shape

    seg = np.maximum(0, (seg - base) / (1 - base))
    segmax = seg.max(axis=0)
    segmax = np.pad(segmax, ((1, 1), (1, 1)))
    data = [
        go.Heatmap(z=segmax, colorscale="Greens", showscale=False, opacity=0.8),
    ]
    if plot_bg and bg.shape[0] > 0:
        bgmax = bg.max(axis=0)
        bgmax = np.pad(bgmax, ((1, 1), (1, 1)))
        data += [
            go.Heatmap(z=bgmax, colorscale="Reds", showscale=False, opacity=0.2),
        ]
    return plot_clip(data, segmax.shape, h + 1, w + 1, scale, margin)


def plot_seg(seg, peaks, mx=None, hsize=20, scale=1, margin=30, label=""):
    def s(i):
        return 1 + i * (size + 1)

    def e(i):
        return (i + 1) * (size + 1)

    peaks = peaks[peaks.kind == "cell"]
    nk = peaks.shape[0]
    seg = seg[:nk]

    size = 2 * hsize + 1
    if mx is None:
        mx = int(np.floor(np.sqrt(nk)))
    my = (nk + mx - 1) // mx
    seg = np.pad(seg, ((0, 0), (hsize, hsize), (hsize, hsize)))
    clip = np.zeros((my * size + (my + 1), mx * size + (mx + 1)), np.float32)

    peaks = peaks[peaks.kind == "cell"]
    ys = np.array(peaks.y)
    xs = np.array(peaks.x)
    for i, (y, x) in enumerate(zip(ys, xs)):
        j, k = divmod(i, mx)
        clip[s(j) : e(j), s(k) : e(k)] = seg[i, y : y + size, x : x + size]

    kwargs = dict(
        colorscale="Greens",
        showscale=False,
    )
    margin = dict(t=margin, b=margin, l=margin, r=margin)
    data = [go.Heatmap(z=clip, **kwargs)]
    return plot_clip(data, clip.shape, size + 1, size + 1, scale, margin)


def plot_calcium(trace, seg, hz, scale=0.3, margin=30, label=""):
    nk, nt = trace.shape
    trace = trace / seg.sum(axis=(1, 2))[:, np.newaxis]
    trace *= -scale
    trace += np.arange(nk)[:, np.newaxis]
    margin = dict(t=margin, b=60 + margin, l=50 + margin, r=margin)
    fig = go.Figure(
        data=[
            go.Scatter(
                x=np.arange(nt) / hz,
                y=trace[k],
                mode="lines",
                line=dict(color="green", width=1),
            )
            for k in range(nk)
        ],
        layout=go.Layout(
            template="none",
            width=nt + margin["l"] + margin["r"],
            height=nk * 3 + margin["t"] + margin["b"],
            margin=margin | dict(autoexpand=False),
            showlegend=False,
        ),
    )
    fig.update_xaxes(
        title="time (sec)",
        showline=False,
    )
    fig.update_yaxes(
        title="cell ID",
        autorange="reversed",
        tickmode="array",
        tickvals=[1] + list(range(100, nk, 100)) + [nk],
        zeroline=False,
    )
    return fig


def plot_spike(spike, hz, diff, time, scale=1, margin=30, label=""):
    spmax = spike.max(axis=1, keepdims=True)
    if time is not None:
        spike = spike[:, slice(*time)]
    spike = spike / np.where(spmax > 0, spmax, 1)
    nk, nt = spike.shape
    margin = dict(t=margin, b=60 + margin, l=50 + margin, r=margin)
    if time is None:
        time = np.arange(nt)
    else:
        time = np.arange(*time)
    fig = go.Figure(
        data=go.Heatmap(
            x=(time - diff) / hz,
            y=np.arange(nk) + 1,
            z=spike,
            colorscale="Reds",
            showscale=False,
        ),
        layout=go.Layout(
            template="none",
            width=nt + margin["l"] + margin["r"],
            height=nk * 3 + margin["t"] + margin["b"],
            margin=margin | dict(autoexpand=False),
            showlegend=False,
        ),
    )
    fig.update_xaxes(
        title="time (sec)",
        showline=False,
    )
    fig.update_yaxes(
        title="cell ID",
        autorange="reversed",
        tickmode="array",
        tickvals=[1] + list(range(100, nk, 100)) + [nk],
        zeroline=False,
    )
    return fig


def xgrid_data(h, w, size):
    return [
        go.Scatter(
            x=[x, x],
            y=[0, h - 1],
            mode="lines",
            line_color="black",
            line_width=1,
        )
        for x in range(0, w, size)
    ]


def ygrid_data(h, w, size):
    return [
        go.Scatter(
            x=[0, w - 1],
            y=[y, y],
            mode="lines",
            line_color="black",
            line_width=1,
        )
        for y in range(0, h, size)
    ]


def plot_clip(data, shape, ysize, xsize, scale, margin):
    h, w = shape
    fig = go.Figure(
        data + xgrid_data(h, w, xsize) + ygrid_data(h, w, ysize),
        layout=go.Layout(
            legend_visible=False,
            width=scale * w + margin["l"] + margin["r"],
            height=scale * h + margin["t"] + margin["b"],
            margin=margin | dict(autoexpand=False),
        ),
    )
    fig.update_xaxes(
        autorange=False,
        range=(-0.5, w - 0.5),
        showticklabels=False,
    )
    fig.update_yaxes(
        scaleanchor="x",
        autorange=False,
        range=(h - 0.5, -0.5),
        showticklabels=False,
    )
    return fig
